__author__ = 'Caitrin'
"""

This file will describe the required components of any class used to provide plotting functionality
It is quite possible that we do not want this to be an abstract base class, as we do not require any functionality..
"""

import abc
import logging
logger = logging.getLogger(__name__)


class Plotter(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def display_something(self):
        """
        this is just a test.
        :return:
        """



# -*- coding: utf-8 -*-
"""
This file defines the TabularDataset Class
"""
import torch
from torch.utils.data import Dataset
import copy
import numpy as np
import pandas as pd
from . import utils as utils
import logging
from itertools import groupby
from sklearn import preprocessing
import time

logger = logging.getLogger(__name__)


# noinspection PyCallingNonCallable
class TabularDataset(Dataset):
    """
    This defines a dataset, subclassed from torch.utils.data.Dataset.

    It uses pd.dataframe as the backend, with utility
    functions.

    Parameters:
            label_column: String
                The name of the label column.
                Provide None if you do not want a target.
            merge_on_columns: list of strings
                Key(s) that specifies which columns to use to uniquely stitch
                dataset (default None)
            index_list: list of strings
                List of columns to make the index of the dataframe
            na_values: The values to convert to NaN when reading from csv
            dataset_dict: keyword parameter, value is dataframe or path string
                pandas dataframe assigned to keyword argument that produces a
                dictionary variable.

        :Example:

        >>> df_test_one: pd.DataFrame({'A': ['A0', 'A1', 'A2', 'A3'],
                                   'B': ['B0', 'B1', 'B2', 'B3'],
                                   'C': ['C0', 'C1', 'C2', 'C3'],
                                   'D': ['D0', 'D1', 'D2', 'D3']},
                                   index=[0, 1, 2, 3]),

        >>> df_test_two: pd.DataFrame({'A': ['A4', 'A5', 'A6', 'A7'],
                                    'B': ['B4', 'B5', 'B6', 'B7'],
                                    'C': ['C4', 'C5', 'C6', 'C7'],
                                    'D': ['D4', 'D5', 'D6', 'D7']},
                                    index=[4, 5, 6, 7])}
        >>> tab_dataset_var = TabularDataset(merge_on_columns=['A'],
                            index_list=['A'], df1=df_test_one, df2=df_test_two)

    """

    def __init__(self, label_column=None, merge_on_columns=None,
                 index_list=None, na_values=None, **dataset_dict):
        """Creates an instance of TabularDataset."""
        for dataset_name in dataset_dict:
            dataset_value = dataset_dict[dataset_name]
            if isinstance(dataset_value, str):
                f_path = dataset_value
                dataset_dict[dataset_name] = pd.read_csv(f_path,
                                                         na_values=na_values,
                                                         index_col=index_list)
            elif not isinstance(dataset_value, pd.DataFrame):
                raise ValueError("Dataset inputs must be either paths \
                                 or DataFrame objects")

        if len(dataset_dict) == 1:
            # TODO: do we want to do anything with this name?
            self.df = dataset_dict[sorted(dataset_dict)[0]]
        else:
            # Not using index list now because we set it before
            self.df = utils.stitch_datasets(dataset_dict, merge_on_columns,
                                            index_list)
            # TODO: check index list doesn't fail if set twice...

        self.label_column = label_column

        self.df = utils.clean_dataframe(self.df)

    def __len__(self):
        """
        The total number of samples.

        Returns:
            The total number of samples: int
        """
        return len(self.df.index)

    def __getitem__(self, idx):
        """
        Generates one sample of data.

        Parameters:
            idx: int
                The index of the data

        Returns:
            The values of the row and the value of the label columns. Xs, Y

        """
        # Where df.drop is used to access the dataframe without
        # the label column, iloc gets the row, then access values and convert

        if self.label_column:
            if self.all_xs is None:
                self.all_xs = self.df.drop(self.label_column,
                              axis=1)
            if self.all_y is None:
                self.all_y = self.df[[self.label_column]]
            xs = torch.from_numpy(self.all_xs.iloc[[idx]].values[0]).float()
            y = self.all_y.iloc[[idx]].values.tolist()[0][0]
            y = torch.tensor(y, dtype=torch.long)
            return xs, y
        else:
            xs = self.df.iloc[[idx]].values.tolist()[0]
            xs = torch.tensor(xs, dtype=torch.float)
            return xs, None

    def merge_dataframe(self, merge_on_columns=None,
                        index_list=None, na_values=None, **dataset_dict):
        """
        Merges additional data into a TabularDataset isntance

        Parameters:
            merge_on_columns: list of strings
                Key(s) that specifies which columns to use to uniquely stitch
                dataset (default None)
            index_list: list of strings
                List of columns to make the index of the dataframe
            na_values: The values to convert to NaN when reading from csv
            dataset_dict: keyword parameter, value is dataframe or path string
                pandas dataframe assigned to keyword argument that produces a
                dictionary variable.

        """
        for dataset in dataset_dict:
            dict_value = dataset_dict[dataset]
            if isinstance(dict_value, str):
                f_path = dict_value
                dataset_dict[dataset] = pd.read_csv(f_path,
                                                    na_values=na_values,
                                                    index_col=index_list)
            elif not isinstance(dict_value, pd.DataFrame):
                raise ValueError("Dataset inputs must be either paths \
                                 or DataFrame objects")

        if len(dataset_dict) == 1:
            df = dataset_dict[sorted(dataset_dict)[0]]
            self.df = utils.stitch_datasets(df_main=self.df,
                                            merge_on_columns=merge_on_columns,
                                            index_list=index_list, new_df=df)
        else:
            if not self.df.empty:
                self.df = utils.stitch_datasets(df_main=self.df,
                                                merge_on_columns=
                                                merge_on_columns,
                                                index_list=index_list,
                                                **dataset_dict)
            else:
                self.df = utils.stitch_datasets(
                    merge_on_columns=merge_on_columns,
                    index_list=index_list,
                    **dataset_dict)

        logger.info("Successfully merged %d datasets", len(dataset_dict))

    def save_dataframe(self, file_path):
        """
        Save the dataframe to a file.

        Parameters:
            file_path: String
                Path to the file where you want your dataframe to be save

        """
        self.df.to_csv(file_path, encoding='utf-8', index=True)
        logger.info("You have saved the dataframe as a csv to %s", file_path)

    def list_all_features(self):
        """
        Lists all features (columns)
        """
        return self.df.columns.values.tolist()

    #  TODO: update to use kwargs.
    def replace_value_in_column(self, column_name, current_values,
                                target_values):
        """
        Replaces one or more values in the given columns.

        Parameters:
            column_name: String
            current_values: List
                Must be existing values
            target_values: List
                Must be valid for pandas dataframe

        """
        if not isinstance(current_values, list):
            current_values = [current_values]
            target_values = [target_values]

        if len(current_values) != len(target_values):
            raise ValueError(
                "Length of current values and target values must be the same")

        self.df[column_name] = self.df[column_name].replace(current_values,
                                                            target_values)

        logger.info("replaced values in %s", column_name)

    def list_all_column_values(self, column_name):
        """
        Return a list of all values in this column.

        Parameters:
            column_name: String
                Name of the column

        """
        return list(getattr(self.df, column_name).unique())

    def print_column_data_types(self):
        """Prints the data types of all columns. """
        print(self.df.dtypes)

    def delete_column(self, column_name):
        """
        Deletes the given column.

        Parameters:
            column_name: String
                The name of the column you want deleted

        """
        self.df = self.df.drop(column_name, axis=1)

        logger.info("You have dropped %s", column_name)

    # TODO: use kwargs
    def create_label_encoding(self, column_name, ordered_values):
        """
        Create label encoding for the given column

        Parameters:
            column_name: String
                The name of the column you want encoded
            ordered_values: List or Dict
                Either an ordered list of possible column values.
                Or a mapping of column value to label value.
                Must include all possible values

        """
        if isinstance(ordered_values, list):
            ordered_values = dict(map(lambda t: (t[1], t[0]), enumerate(
                ordered_values)))
        elif not isinstance(ordered_values, dict):
            raise ValueError("Must be either a list or dictionary")

        column_vals = ordered_values.keys()

        if len(column_vals) != len(set(column_vals)):
            raise ValueError("Ordered_value_list contains non-unique values")

        if set(column_vals) != set(self.list_all_column_values(column_name)):
            raise ValueError("Not all column values are included")

        self.df[column_name] = getattr(self.df, column_name).map(
            ordered_values)

        logger.info("Successfully remapped %s", column_name)

    def create_one_hot_encoding(self, column_name, prefix_sep="@"):
        """
        Create one-hot encoding for the given column.

        Parameters:
            column_name: String
                The name of the column you want to one-hot encode
            prefix_sep String default("@")
                The prefix used when creating a one-hot encoding

        """
        # TODO: ensure dummy_na =False is what you want
        self.df = pd.get_dummies(self.df, columns=[column_name],
                                    prefix_sep=prefix_sep)
        logger.info("Successfully encoded %s", column_name)

    # if a use case presents itself column_name could easily become a list
    def reverse_create_one_hot_encoding(self, column_name=None,
                                        prefix_sep="@"):
        """
        Undo the creation of one-hot encodings, if prefix_sep was used
        to create one-hot encodings and nowhere else. If a column_name is
        provided, only that column will be reverse-encoded, otherwise all will
        be reverse-encoded.

        Parameters
            column_name: String
                The name of the column to reverse the one-hot encoding for
            prefix_sep: String default("@")
                The prefix used when creating a one-hot encoding

        """
        result_series = {}

        if column_name:
            if prefix_sep in column_name:
                considered_column_list = [column_name]
            else:
                raise ValueError("Prefix not found in column_name")
            non_dummy_cols = [col for col in self.df.columns
                              if col not in considered_column_list]
        else:
            considered_column_list = self.df.columns
            non_dummy_cols = [col for col in self.df.columns
                              if prefix_sep not in col]

        # Find dummy columns and build pairs (category, category_value)
        dummy_tuples = [(col.split(prefix_sep)[0], col) for col in
                        considered_column_list if prefix_sep in col]

        for dummy, cols in groupby(dummy_tuples, lambda item: item[0]):
            dummy_df = self.df[[col[1] for col in cols]]

            # Find max value among columns
            max_columns = dummy_df.idxmax(axis=1)

            # Remove category_ prefix
            result_series[dummy] = max_columns.apply(lambda item:
                                                     item.split(prefix_sep)[1])

        # Copy non-dummy columns over.
        for col in non_dummy_cols:
            result_series[col] = self.df[col]

        self.df = pd.DataFrame(result_series)

        logger.info("Successfully converted %d \
                    columns back from dummy format.", len(dummy_tuples))

    def identify_null(self, threshold):
        """
        Return columns where there is at least threshold percent of
        null values.

        Parameters:
            threshold: Float
                A number between 0 and 1, representing proportion of null
                values.

        Returns:
            cols: List
                A list of those columns with threshold percentage null values

        """
        if threshold >= 1 or threshold <= 0:
            raise ValueError(
                "Threshold needs to be a proportion between 0 and 1 \
                (exclusive)")
        num_threshold = ( (1- threshold) * len(self))
        # thresh is "Require that many non-NA values."
        tmp = self.df.dropna(thresh=num_threshold, axis=1)
        cols = list(set(self.df.columns).difference(set(tmp.columns)))
        return cols

    def identify_unique(self, threshold):
        """
        Returns columns that do not have have at least threshold number of
        values. If a column has 9 values and the threshold is 9, that column
        will not be returned.

        Parameters:
            threshold: The minimum number of values needed.
                Must be greater than 1. Not between 0 and 1, but rather
                represents the number of values

        Returns:
            column_list: list
                The list of columns having threshold number of values

        """
        obj_types = {col: set(map(type, self.df[col])) for col in
                     self.df.columns}

        column_list = []
        for col in self.df.columns:
            if len(obj_types[col]) > 1:
                logger.warning("Column: {} has mixed datatypes, this may"
                               "interfere with an accurate identification"
                               "of mixed values: i.e. you may have 1 and '1'"
                               .format(col))
            if len(self.df[col].unique()) < threshold:
                column_list.append(col)
        return column_list

    def identify_unbalanced_columns(self, threshold, non_numeric=True):
        """
        This returns columns that are highly unbalanced.
        Those that have a disproportionate amount of one value.

        Parameters:
            threshold: Float
                Proportion needed to define unbalanced, between 0 and 1
                0 is a lesser proportion of the one value, 
                (less imbalanced)
            non_numeric: Boolean
                Whether non-numeric columns are also considered.

        Returns:
            column_list: List
                The list of column names

        """
        if non_numeric:
            columns = list(self.df.columns)
        else:
            columns = list(self.df.select_dtypes(include=np.number))
        column_list = []
        for col in columns:
            # Check amount of null values, because if column is entirely null,
            # max won't work.
            num_of_null = self.df[col].isnull().sum()
            if num_of_null != len(self):
                col_maj = (max(self.df[col].value_counts()) /
                           self.df[col].value_counts().sum())
                if col_maj >= threshold:
                    column_list.append(col)
        return column_list

    def identify_highly_correlated(self, threshold):
        """
        Identify columns that are highly correlated with one-another.

        Parameters:
            threshold: Between 0 (weakest correlation) and 1
                (strongest correlation).
                Minimum amount of correlation necessary to be identified.

        Returns:
            column list: List of tuples
                The correlation values above threshold

        """
        column_list = set()
        features_correlation = self.df.corr().abs()
        corr_pairs = features_correlation.unstack()
        for index, val in corr_pairs.items():
            if val > threshold and (index[0] != index[1]):
                column_list.add((index, val))
        return column_list

    def identify_low_variance(self, threshold):
        """
        Identify those columns that have low variance

        Parameters:
            threshold: Float
                Between 0 an 1.
                Maximum amount of variance necessary to be identified as low
                variance.

        Returns:
            variance_dict: Dict
                A dictionary of column names, with the value being their
                variance.

        """
        dct_low_var = {}
        scaler = preprocessing.MinMaxScaler()
        for col in self.df.columns:
            if self.df[col].dtype in ['float64', 'int64', 'float32', 'int32']:
                col_float_array = self.df[col].dropna().values.astype(float)
                reshaped_col_float_array = col_float_array.reshape(-1, 1)
                scaled_col = scaler.fit_transform(reshaped_col_float_array)
                col_var = scaled_col.var()
                if col_var <= threshold:
                    dct_low_var[col] = col_var
        return dct_low_var

    def convert_all_categorical_binary(self, list_only=False,
                                       exception_columns=None):
        """Recodes all columns with only two values as ones and zeros, float
        valued.

        This should only be used once you have a final dataset in case values
        are not actually binary. Useful for dealing with all the variations of
        YES/NO ,yEs/nO etc.

        Parameters:
             list_only: boolean
                only return a list of columns for which this would
                apply, do not actually do the transformation.
                If false operation is performed in place.
            exception_columns: list
                list of column names you do not wish to convert.

        Returns:
            list or None
            list if list_only if true, nothing otherwise.
        """

        # recoding binary valued columns as ones and zeros

        binary_cols = [(col, self.df[col].value_counts().index) for col
                       in self.df.columns if
                       len(self.df[col].value_counts()) == 2]

        if list_only:
            return binary_cols

        for col, index in binary_cols:
            di = {index[0]: 1.0, index[1]: 0.0}
            try:
                self.df = self.df.replace({col: di})
            except:
                continue
            self.df[col] = self.df[col].astype(np.float64)

    # future improvements could come from
    # https://github.com/pytorch/text/blob/master/torchtext/data/dataset.py
    # noinspection PyUnusedLocal
    def split(self, split_ratio=0.7, stratified=False, stratum_column=None):
        """
        Create train-test(-validation) splits from the instance's examples.
        Function signature borrowed from torchtext in an effort to maintain
        consistency
        https://github.com/pytorch/text/blob/master/torchtext/data/dataset.py
        also partially modified from
        https://stackoverflow.com/questions/38250710/
        how-to-split-data-into-3-sets-train-validation-and-test

        Parameters:
            split_ratio: Float
                a number [0, 1] denoting the amount
                of data to be used for the training split (rest is used for
                validation), or a list of numbers denoting the relative sizes
                of train, test and valid splits respectively.
                If the relative size for valid is missing, only the
                train-test split is returned.
                Default is 0.7 (for th train set).
            stratified: Boolean
                whether the sampling should be stratified.
                    Default is False.
            stratum_column: String
                name of the examples column stratified over.
                Default is 'label_column'

        Returns:
            datasets: Tuple of TabularDatasets
                Datasets for train, test, validation
                 splits in that order, if the splits are provided.

        """

        train_ratio, test_ratio, validation_ratio = utils.check_split_ratio(
            split_ratio)

        train_indices = []
        test_indices = []
        validation_indices = []

        if stratified:
            if not stratum_column:
                stratum_column = self.label_column
            else:
                if stratum_column not in self.df.columns:
                    raise ValueError("Invalid strata column name")

            grps = self.df.groupby(stratum_column)
            train_index, test_index, val_index = [], [], []
            for key, grp in grps:
                group_train, group_test, group_val = \
                    utils.rationed_split(grp, train_ratio, test_ratio,
                                         validation_ratio)

                train_indices += list(group_train)
                test_indices += list(group_test)
                validation_indices += list(group_val)

        else:

            train_indices, test_indices, validation_indices = \
                utils.rationed_split(self.df, train_ratio,
                                     test_ratio, validation_ratio)

        train = TabularDataset(train=self.df.loc[train_indices],
                               label_column=self.label_column)
        test = TabularDataset(test=self.df.loc[test_indices],
                              label_column=self.label_column)

        if validation_ratio:
            validation = TabularDataset(val=self.df.loc[validation_indices],
                                        label_column=self.label_column)
            return train, test, validation
        else:
            return train, test



from vulcanai.datasets.utils import stitch_datasets
import pytest
import pandas as pd
import numpy as np


class TestStitchDataset:
    @pytest.fixture
    def my_test_dataset_one(self):
        dct_dfs = {'df_test_one': pd.DataFrame({'A': ['A0', 'A1', 'A2', 'A3'], 'B': ['B0', 'B1', 'B2', 'B3'],
                                                'C': ['C0', 'C1', 'C2', 'C3'], 'D': ['D0', 'D1', 'D2', 'D3']},
                                               index=[0, 1, 2, 3]),
                   'df_test_two': pd.DataFrame({'A': ['A4', 'A5', 'A6', 'A7'], 'B': ['B4', 'B5', 'B6', 'B7'],
                                                'C': ['C4', 'C5', 'C6', 'C7'], 'D': ['D4', 'D5', 'D6', 'D7']},
                                               index=[4, 5, 6, 7])}
        return dct_dfs

    @pytest.fixture
    def my_test_dataset_two(self):
        dct_dfs = {'df_test_one': pd.DataFrame({'name': ['Jane', 'John', 'Jesse', 'Jane', 'John', 'Jesse'],
                                                'age': [23, 25, 26, np.nan, np.nan, np.nan]},
                                               index=[0, 1, 2, 3, 4, 5]),
                   'df_test_two': pd.DataFrame({'name': ['Jane', 'John', 'Jesse'],
                                                'state': ['CA', 'WA', 'OR']},
                                               index=[0, 1, 2])}
        return dct_dfs

    @pytest.fixture
    def my_test_dataset_three(self):
        dct_dfs = {'df_test_one': pd.DataFrame({'name': ['Jane', 'John', 'Jesse', 'Jane', 'John', 'Jesse', 'Jane'],
                                                'age': [23, 25, 26, np.nan, np.nan, np.nan, 23],
                                                'dob': ['09-18-1995', '10-18-1993', '06-18-1992', np.nan, np.nan,
                                                        np.nan,
                                                        '05-23-1995']},
                                               index=[0, 1, 2, 3, 4, 5, 6]),
                   'df_test_two': pd.DataFrame({'name': ['Jane', 'John', 'Jesse', 'Jane'],
                                                'dob': ['09-18-1995', '10-18-1993', '06-18-1992', '05-23-1995'],
                                                'state': ['CA', 'WA', 'OR', 'AZ']},
                                               index=[0, 1, 2, 3])}
        return dct_dfs

    @pytest.fixture
    def my_test_dataset_four(self):
        dct_dfs = {'df_test_one': pd.DataFrame({'name': ['Jane', 'John', 'Jesse', 'Jane'],
                                                'age': [23, 25, 26, 23],
                                                'dob': ['09-18-1995', '10-18-1993', '06-18-1992', '05-23-1995'],
                                                'visit_date': ['11/29/2018', '11/29/2018', '11/29/2018', '11/29/2018'],
                                                'visit_location': ['HI', 'HI', 'HI', 'HI']},
                                               index=[0, 1, 2, 3]),
                   'df_test_two': pd.DataFrame({'name': ['John', 'Jesse'],
                                                'age': [25, 26],
                                                'dob': ['10-18-1993', '06-18-1992'],
                                                'visit_date': ['09/12/2018', '12/20/2017'],
                                                'visit_location': ['CA', 'AZ']},
                                               index=[0, 1])}
        return dct_dfs

    def test_no_merge_on_columns(self, my_test_dataset_one):
        # MOC (merge on columns)
        df_no_moc_results = pd.DataFrame({'A': ['A0', 'A1', 'A2', 'A3', 'A4', 'A5', 'A6', 'A7'],
                                          'B': ['B0', 'B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7'],
                                          'C': ['C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7'],
                                          'D': ['D0', 'D1', 'D2', 'D3', 'D4', 'D5', 'D6', 'D7']},
                                         index=[0, 1, 2, 3, 4, 5, 6, 7])

        stitch_dataset_results = stitch_datasets(merge_on_columns=None, **my_test_dataset_one)
        pd.testing.assert_frame_equal(stitch_dataset_results, df_no_moc_results)

    def test_single_merge_on_columns(self, my_test_dataset_two):
        df_single_moc_results = pd.DataFrame({'name': ['Jane', 'John', 'Jesse'],
                                              'age': [23, 25, 26],
                                              'state': ['CA', 'WA', 'OR']},
                                             index=[0, 1, 2])
        stitch_dataset_results = stitch_datasets(merge_on_columns=['name'], **my_test_dataset_two)

        # Assert_frame_equal checks order of columns; therefore, sort_index by columns when checking. If dataframes are
        # same, they should sort the same way.
        pd.testing.assert_frame_equal(stitch_dataset_results.sort_index(axis=1), df_single_moc_results.sort_index(axis=1),
                                      check_dtype=False)

    def test_two_merge_on_columns(self, my_test_dataset_three):
        df_two_moc_results = pd.DataFrame({'name': ['Jane', 'John', 'Jesse', 'Jane'],
                                              'age': [23, 25, 26, 23],
                                              'dob': ['09-18-1995', '10-18-1993', '06-18-1992', '05-23-1995'],
                                              'state': ['CA', 'WA', 'OR', 'AZ']},
                                             index=[0, 1, 2, 6])
        stitch_dataset_results = stitch_datasets(merge_on_columns=['name', 'dob'], **my_test_dataset_three)

        # Assert_frame_equal checks order of columns; therefore, sort_index by columns when checking. If dataframes are
        # same, they should sort the same way.
        pd.testing.assert_frame_equal(stitch_dataset_results.sort_index(axis=1), df_two_moc_results.sort_index(axis=1),
                                      check_dtype=False)

    def test_three_merge_on_columns(self, my_test_dataset_four):
        df_three_moc_results = pd.DataFrame({'name': ['Jane', 'John', 'Jesse', 'Jane', 'John', 'Jesse'],
                                             'age': [23, 25, 26, 23, 25, 26],
                                             'dob': ['09-18-1995', '10-18-1993', '06-18-1992', '05-23-1995',
                                                     '10-18-1993', '06-18-1992'],
                                             'visit_date': ['11/29/2018', '11/29/2018', '11/29/2018', '11/29/2018',
                                                            '09/12/2018', '12/20/2017'],
                                             'visit_location': ['HI', 'HI', 'HI', 'HI', 'CA', 'AZ']},
                                            index=[0, 1, 2, 3, 4, 5])

        stitch_dataset_results = stitch_datasets(merge_on_columns=['name', 'dob', 'visit_date'], **my_test_dataset_four)

        # Assert_frame_equal checks order of columns; therefore, sort_index by columns when checking. If dataframes are
        # same, they should sort the same way.
        pd.testing.assert_frame_equal(stitch_dataset_results.sort_index(axis=1), df_three_moc_results.sort_index(axis=1),
                                      check_dtype=False)


import logging.config
import os
import torch
import random
import numpy as np


DEFAULT_RANDOM_SEED = 42

log_config_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                               'logging.conf')

if not os.path.isfile(log_config_path):
    raise IOError

log_output_path = "logfile.log"

logging.config.fileConfig(log_config_path,
                          disable_existing_loggers=False,
                          defaults={'logfilename': log_output_path})

logger = logging.getLogger(__name__)


def set_global_seed(seed_value):
    """
    Sets all the random seeds, including for torch, GPU, numpy and python.

    Parameters:
        seed_value : int
            The random seed value.

    """
    random.seed(seed_value)
    np.random.seed(seed_value)  # cpu vars
    torch.manual_seed(seed_value)  # cpu  vars
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed_value)


set_global_seed(DEFAULT_RANDOM_SEED)




# -*- coding: utf-8 -*-
"""Defines the basenetwork class."""
# Core imports
import abc
import torch
from torch import nn

# Vulcan imports
from .layers import *
from .utils import set_tensor_device

from .metrics import Metrics
from ..plotters.visualization import display_record, get_save_path

# Generic imports
import pydash as pdash
from tqdm import tqdm, trange
from datetime import datetime
import logging
import os
from os import environ
import pickle
import numpy as np

import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns

import warnings

warnings.filterwarnings("ignore")

sns.set(style='dark')
logger = logging.getLogger(__name__)


class BaseNetwork(nn.Module):
    """
    Defines the BaseNetwork object.

    Parameters:
        name : str
            The name of the network. Used when saving the file.
        config : dict
            The configuration of the network module, as a dict.
        in_dim : tuple
            The input dimensions of the network. Not required to specify when the
            network has input_networks.
        save_path : str
            The name of the file to which you would like to save this network.
        input_networks : list of BaseNetwork
            A network object provided as input.
        num_classes : int or None
            The number of classes to predict. In the binary
            case always specify 2 classes no matter the shape of the data. A
            value of 1 is used for predicting a single continuous value.
        activation : torch.nn.Module
            The desired activation function for use in the network.
        pred_activation : torch.nn.Module
            The desired activation function for use in the prediction layer.
        optim_spec : dict
            A dictionary of parameters for the desired optimizer.
        lr_scheduler : torch.optim.lr_scheduler
            A callable torch.optim.lr_scheduler
        early_stopping : str or None
            So far just 'best_validation_error' is implemented.
        early_stopping_patience: integer
            Number of validation iterations of decreasing loss
            (note -not necessarily every epoch!
            before early stopping is applied.
        early_stopping_metric: string
            Either "loss" or "accuracy" are implemented.
        criter_spec : dict
            criterion specification with name and all its parameters.
        device : str or torch.device
            Sets the network module to the relevant device. If cuda available
            in the host console, "cuda:0" will be run which can be overriden.

    Returns:
        network : BaseNetwork

    """

    def __init__(self, name, config, in_dim=None, save_path=None,
                 input_networks=None, num_classes=None,
                 activation=nn.ReLU(), pred_activation=None,
                 optim_spec={'name': 'Adam', 'lr': 0.001},
                 lr_scheduler=None, early_stopping=None,
                 early_stopping_patience=None,
                 early_stopping_metric = "accuracy",
                 criter_spec=nn.CrossEntropyLoss(),
                 device="cuda:0"):
        """Define, initialize, and build the BaseNetwork."""
        super(BaseNetwork, self).__init__()

        assert isinstance(name, str)
        self._name = name
        self._config = config
        self._save_path = save_path

        # Turn into list if not list
        if input_networks and not isinstance(input_networks, list):
            input_networks = [input_networks]

        if input_networks:
            self.input_networks = nn.ModuleDict()
            for in_net in input_networks:
                self._add_input_network(in_net)
        else:
            self.input_networks = input_networks

        self._set_final_layer_parameters(pred_activation=pred_activation,
                                         criter_spec=criter_spec)

        self._num_classes = num_classes

        self._lr_scheduler = lr_scheduler
        self._early_stopping = early_stopping
        self._early_stopping_patience = early_stopping_patience
        self._early_stopping_metric = early_stopping_metric

        if self._num_classes:
            self.metrics = Metrics()

        self.optim = None
        self.criterion = None

        self.epoch = 0

        self.record = dict(
            epoch=[],
            train_error=[],
            train_accuracy=[],
            validation_error=[],
            validation_accuracy=[]
        )

        if in_dim:
            if isinstance(in_dim, int):
                self.in_dim = tuple([in_dim])
            else:
                self.in_dim = in_dim
        else:
            if input_networks is None:
                raise ValueError("BaseNetwork must have either in_dim or \
                                input_networks")

        # Creates the Network, and re-writes the self.in_dim
        self._create_network(
            activation=activation,
            pred_activation=pred_activation)

        # Compute self.out_dim of the network
        self.out_dim = self._get_out_dim()

        self.device = device
        self._optim_spec = optim_spec
        self._criter_spec = criter_spec

    def _add_input_network(self, in_network):
        """
        Add a new network to  an input for this network.

        New input network will exist at the end of the current set of
        input_networks which will need to correspond with .

        Parameters:
            in_network : BaseNetwork
                A network to append to the set of self.input_networks.

        Returns:
            None

        """
        assert isinstance(in_network, BaseNetwork)
        assert in_network is not self
        if not self.input_networks:
            self.input_networks = nn.ModuleDict()
        else:
            assert isinstance(self.input_networks, nn.ModuleDict)
        self.input_networks[in_network.name] = in_network
        self.in_dim = self._get_in_dim()

    def _set_final_layer_parameters(self, pred_activation, criter_spec):
        """
        Sets and checks the parameters used in the final output layer.

        Final transform is needed in forward pass in the case of
        nn.CrossEntropyLoss as this class combines nn.NLLLoss and softmax,
        meaning the outputs are not softmax transformed.

        Parameters:
            pred_activation : torch.nn.Module
                The desired activation function for the prediction layer.
            criter_spec : dict
                criterion specification with name and all its parameters.

        """
        self._final_transform = None

        if isinstance(criter_spec, nn.CrossEntropyLoss):
            if pred_activation:
                raise ValueError("The nn.CrossEntropyLoss class combines  \
                            nn.NLLLoss and softmax for improved efficiency, \
                            you cannot set pred_activation when \
                            criter_spec is set to an instance of this class. \
                            Set pred_activation to none or change criter_spec."
                                 )
            self._final_transform = nn.Softmax(dim=1)

    @abc.abstractmethod
    def _merge_input_network_outputs(self, inputs):
        """Abstract method used to define how to handle multi-inputs."""
        raise NotImplementedError

    def extra_repr(self):
        """Set the extra representation of the module."""
        return '(device): torch.'+self.device.__repr__()

    @torch.no_grad()
    def _get_out_dim(self):
        """
        Return the network output shape.

        Perform a single forward pass using made-up data.

        Returns:
            shape : tuple
                The output shape of the network.

        """
        if self.network:
            out = self.network(torch.ones([1, *self.in_dim]))
            return tuple(out.shape[1:])
        else:
            return None

    @torch.no_grad()
    def _get_in_dim(self):
        """
        Return the network input shape.

        Perform a single forward pass through all the input networks
        and merge together to get input shape of this network.

        Returns:
            shape : tuple
                The input shape of the network.

        """
        # Create empty input tensors
        in_tensors = []
        for in_net in self.input_networks.values():
            in_tensors.append(torch.ones([1, *in_net.out_dim]))
        output = self._merge_input_network_outputs(in_tensors)
        return tuple(output.shape[1:])

    @property
    def device(self):
        """
        Return the relevant device associalted with network module.

        Returns:
            device : torch.device
                Relevant device associalted with the network module.

        """
        return next(self.network.parameters()).device

    @device.setter
    def device(self, device):
        """
        Network device setter.

        If the user specifies invalid device id, raises
        RuntimeError for invalid device ordinal.

        Parameters:
            device : str or torch.device
                The device to transfer network to.

        """
        device = torch.device(device if torch.cuda.is_available() else 'cpu')
        if self.network:
            # nonblocking doesn't seem to matter here
            self.network.to(device=device)
        if self.optim:
            for state in self.optim.state.values():
                for k, v in state.items():
                    if torch.is_tensor(v):
                        state[k] = set_tensor_device(v, device=device)

    @property
    def is_cuda(self):
        """
        Return  boolean about whether the network is on cuda device (i.e gpu).

        Returns:
            is_cuda : boolean
                Specifies whether the network is on gpu or not.

        """
        return next(self.network.parameters()).is_cuda

    @property
    def name(self):
        """
        Return the name.

        Returns:
            name : string
                The name of the network.

        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def save_path(self):
        """Return the save path of the network.

        Returns:
            save_path : string
                The save path of the network.

        """
        return self._save_path

    @save_path.setter
    def save_path(self, value):
        if not value:
            self._save_path = "{}_{date:%Y-%m-%d_%H_%M_%S}/".format(
                self.name, date=datetime.now())
        else:
            self._save_path = value

    @property
    def lr_scheduler(self):
        """
        Return the network lr_scheduler.

        Returns:
            lr_scheduler : torch.optim.lr_scheduler

        """
        return self._lr_scheduler

    @lr_scheduler.setter
    def lr_scheduler(self, value):
        self._lr_scheduler = value

    @property
    def early_stopping(self):
        """
        Return the stopping rule.

        Returns:
            stopping_rule : str
                The stoping rule

        """
        return self._early_stopping

    @early_stopping.setter
    def early_stopping(self, value):
        self._early_stopping = value

    @property
    def early_stopping_patience(self):
        """
        Return the early stopping patience.

        Returns:
            early_stopping_patience : int
                The early stopping patience

        """
        return self._early_stopping_patience

    @early_stopping_patience.setter
    def early_stopping_patience(self, value):
        self._early_stopping_patience = value



    @property
    def early_stopping_metric(self):
        """
        Return the early stopping netric.

        Returns:
            early_stopping_metric : string
                The early stopping metric

        """
        return self._early_stopping_metric

    @early_stopping_metric.setter
    def early_stopping_metric(self, value):
        self._early_stopping_metric = value

    @property
    def criter_spec(self):
        """
        Return the criterion specification.

        Returns:
            _criter_spec : dict
                The criterion specification.

        """
        return self._criter_spec

    @criter_spec.setter
    def criter_spec(self, value):
        self._criter_spec = value

    def get_layers(self):
        """
        Return an ordered dict of all modules in this network (layers).

        Returns:
            layers : OrderedDict()

        """
        return self._modules

    def get_weights(self):
        """
        Return a dictionary containing a whole state of the module.

        Returns:
            weights : dict
                A dictionary containing a whole state of the module.

        """
        return self.state_dict()

    @abc.abstractmethod
    def _create_network(self, **kwargs):
        """
        Define the network. Abstract method that needs to be overridden.

        Returns:
            None

        """
        self.network = None
        pass

    def freeze(self, apply_inputs=False):
        """
        Freeze network weights so training doesn't modify them.

        Parameters:
            apply_inputs : boolean
                Whether to freeze all input networks recursively

        Returns:
            None

        """
        self._toggle_freeze(freeze=True, apply_inputs=apply_inputs)

    def unfreeze(self, apply_inputs=False):
        """
        Unfreeze network weights so training does modify them.

        Parameters:
            apply_inputs : boolean
                Whether to unfreeze all input networks recursively

        Returns:
            None

        """
        self._toggle_freeze(freeze=False, apply_inputs=apply_inputs)

    def _toggle_freeze(self, freeze, apply_inputs):
        # Freeze core network parameters
        for params in self.network.parameters():
            # If freeze is True, set requires_grad to False
            # If freeze is False, set requires_grad to True
            params.requires_grad_(not freeze)
        # Recursively toggle freeze on
        if apply_inputs and self.input_networks:
            for in_net in self.input_networks.values():
                in_net._toggle_freeze(
                    freeze=freeze,
                    apply_inputs=apply_inputs)

    def _init_optimizer(self, optim_spec):
        optim_class = getattr(torch.optim, optim_spec["name"])
        optim_spec = pdash.omit(optim_spec, "name")
        return optim_class(self.parameters(), **optim_spec)

    @staticmethod
    def _init_criterion(criterion_spec):
        return criterion_spec

    def _init_trainer(self):
        self.optim = self._init_optimizer(self._optim_spec)
        # TODO: Use logger to describe if the optimizer is changed.
        self.criterion = self._init_criterion(self._criter_spec)

    def _assert_same_devices(self, comparison_device=None):
        """
        Will check if all incoming networks are on the same device.

        Raises specific error about which device the networks need to
        be re-assigned to.

        Specific for when calling fit becuase of grad calculation across
        several devices not compatible with optimizer. Temporary until
        replaced with nn.DataParallel or better multi-gpu implmentation.

        Parameters:
            comparison_device : str or torch.device
                The device to compare current device to.

        """
        if comparison_device is None:
            comparison_device = self.device
        incompatible_collector = {}
        if self.input_networks:
            for net_name, net in self.input_networks.items():
                if net.input_networks:
                    net._assert_same_devices(comparison_device)
                if net.device != comparison_device:
                    incompatible_collector[net_name] = net.device
        if incompatible_collector:
            raise ValueError(
                "The following input networks' devices do not "
                "match deepest network's device '{}':\n{}".format(
                    comparison_device,
                    incompatible_collector))

    # modified from
    # https://github.com/Bjarten/early-stopping-pytorch/blob/master
    # /pytorchtools.py
    class EarlyStopping:
        """Early stops the training if validation loss doesn't improve after
        a given number of epochs.
        Args:
            patience: int default 7
                How many epochs to wait after last time
                validation loss improved.
            verbose: bool default False
                If True, prints a message for each
                validation loss improvement.
        """

        def __init__(self, patience=2, verbose=False):
            """ Initialize early stopping """
            self.patience = patience
            self.verbose = verbose
            self.counter = 0
            self.best_score = None
            self.early_stop = False
            self.val_loss_min = np.Inf
            self.best_model = None
            self.save_path = ""

        def __call__(self, score, model):

            # necessary with current code
            if score:
                score = float(score)

            if not score:
                logger.info('EarlyStopping counter not incremented, score'
                            'was not calculated this epoch')

            elif score and not np.isnan(score) and self.best_score is None:
                self.best_score = score
                self.save_checkpoint(score, model)

            # never save a model that has nan as a score, but count
            # it towards total (i.e. don't nan to infinity, exit early).
            elif score <= self.best_score or np.isnan(score):
                self.counter += 1
                logger.info('EarlyStopping counter: {} out of {}'.format(
                    self.counter, self.patience
                ))
                if self.counter >= self.patience:
                    self.early_stop = True

            else:
                self.best_score = score
                self.save_checkpoint(score, model)
                self.counter = 0

        def save_checkpoint(self, val_loss, model):
            '''Saves model when validation loss decrease.'''
            if self.verbose:
                logger.info(
                    f'Validation loss decreased ({self.val_loss_min:.6f} --> '
                    f'{val_loss:.6f}).  Saving model ...')
            self.save_path = model.save_model()
            self.val_loss_min = val_loss

    def fit(self, train_loader, val_loader, epochs,
            retain_graph=None, valid_interv=4, plot=False, save_path=None):
        """
        Train the network on the provided data.

        Parameters:
            train_loader : DataLoader
                The DataLoader object containing the training data.
            val_loader : DataLoader
                The DataLoader object containing the validation data.
            epochs : int
                The number of epochs to train for.
            retain_graph : {None, True, False}
                Whether retain_graph will be true when .backwards is called.
            valid_interv : int
                Specifies the period of epochs before validation calculation.
            plot : boolean
                Whether or not to plot training metrics in real-time.
            save_path : str
                Path to save graphics at

        Returns:
            None

        """
        # Check all networks are on same device.
        self._assert_same_devices()

        # In case there is already one, don't overwrite it.
        # Important for not removing the ref from a lr scheduler
        if self.optim is None:
            self._init_trainer()

        early_stopping = self.EarlyStopping(patience=
                                            self.early_stopping_patience,
                                            verbose=True)

        try:
            if plot:
                fig_number = plt.gcf().number + 1 if plt.fignum_exists(1) else 1
                plt.show()

            if save_path:
                if save_path.endswith('/'):
                    save_path = save_path + self.name + '_'
                else:
                    save_path = save_path + '/' + self.name + '_'
                save_path = get_save_path(save_path, vis_type='train')
            iterator = trange(epochs, desc='Epoch: ')

            for epoch in iterator:

                train_loss, train_acc = self._train_epoch(train_loader,
                                                          retain_graph)

                valid_loss = valid_acc = None
                if epoch % valid_interv == 0:
                    valid_loss, valid_acc = self._validate(val_loader)

                if self.lr_scheduler:
                    self.lr_scheduler.step(epoch=epoch)

                if self.early_stopping:
                    if self.early_stopping_metric == "loss":
                        if valid_loss:
                            valid_loss_neg = -1 * valid_loss
                        early_stopping(valid_loss_neg, self)
                    elif self.early_stopping_metric == "accuracy":
                        early_stopping(valid_acc, self)
                    else:
                        raise ValueError("Not a valid stopping metric")

                    if early_stopping.early_stop:
                        logger.info("Early stopping")
                        # restores everything to the earlier saved version
                        self.__dict__.update(self.load_model(
                            early_stopping.save_path).__dict__)
                        # for tqdm
                        iterator.close()
                        break

                # reset from None so that a distinction can be made
                # between no value being collected
                if not valid_loss:
                    valid_loss = np.nan
                if not valid_acc:
                    valid_acc = np.nan

                tqdm.write(
                    "\n Epoch {}:\n"
                    "Train Loss: {:.6f} | Test Loss: {:.6f} |"
                    "Train Acc: {:.4f} | Test Acc: {:.4f}".format(
                        self.epoch,
                        train_loss,
                        valid_loss,
                        train_acc,
                        valid_acc))

                self.record['epoch'].append(self.epoch)
                self.record['train_error'].append(train_loss)
                self.record['train_accuracy'].append(train_acc)
                self.record['validation_error'].append(valid_loss)
                self.record['validation_accuracy'].append(valid_acc)

                if plot:
                    plt.ion()
                    plt.figure(fig_number)
                    display_record(record=self.record, save_path=save_path)

                self.epoch += 1

        except KeyboardInterrupt:
            logger.warning(
                "\n\n**********KeyboardInterrupt: "
                "Training stopped prematurely.**********\n\n")

    def _train_epoch(self, train_loader, retain_graph):
        """
        Trains the network for 1 epoch.

        Parameters:
            train_loader : DataLoader
                The DataLoader object containing the dataset to train on.

        Returns:
            (train_loss, train_accuracy) : (float, float)
                Returns the train loss and accuracy.

        """
        self.train()  # Set model to training mode

        train_loss_accumulator = 0.0
        train_accuracy_accumulator = 0.0
        pbar = trange(len(train_loader.dataset), desc='Training.. ')

        for data, targets in train_loader:
            data = set_tensor_device(data, device=self.device)
            targets = set_tensor_device(targets, device=self.device)

            # Forward + Backward + Optimize
            predictions = self(data)
            train_loss = self.criterion(predictions, targets)
            train_loss_accumulator += train_loss.item()

            self.optim.zero_grad()
            train_loss.backward(retain_graph=retain_graph)
            self.optim.step()

            # here can add more options in the future if other metrics
            # will be used
            average = None

            if self._num_classes == 1:
                metric = "mse"
            else:
                metric = "accuracy"

            if self._final_transform:
                predictions = self._final_transform(predictions)

            # will be fixed in the future
            train_accuracy_accumulator += self.metrics.get_score(
                targets=targets,
                predictions=predictions,
                metrics=metric,
                average=average)[metric]

            pbar.update(train_loader.batch_size)
        pbar.close()

        train_loss = train_loss_accumulator * \
            train_loader.batch_size / len(train_loader.dataset)
        train_accuracy = train_accuracy_accumulator * \
            train_loader.batch_size / len(train_loader.dataset)

        return train_loss, train_accuracy

    @torch.no_grad()
    def _validate(self, val_loader):
        """
        Validate the network on the validation data.

        Parameters:
            val_loader : DataLoader
                The DataLoader object containing the dataset to evaluate on

        Returns:
            (val_loss, val_accuracy) : (float, float)
                Returns the validation loss and accuracy

        """
        self.eval()  # Set model to evaluate mode

        val_loss_accumulator = 0.0
        val_accuracy_accumulator = 0.0
        pbar = trange(len(val_loader.dataset), desc='Validating.. ')

        for data, targets in val_loader:

            data = set_tensor_device(data, device=self.device)
            targets = set_tensor_device(targets, device=self.device)

            predictions = self(data)
            validation_loss = self.criterion(predictions, targets)
            val_loss_accumulator += validation_loss.item()

            # here can add more options in the future if other metrics
            # will be used
            average = None

            if self._num_classes == 1:
                metric = "mse"
            else:
                metric = "accuracy"

            if self._final_transform:
                predictions = self._final_transform(predictions)

            # will be fixed in the future
            val_accuracy_accumulator += self.metrics.get_score(
                targets=targets,
                predictions=predictions,
                metrics=metric,
                average=average)[metric]

            pbar.update(val_loader.batch_size)
        pbar.close()

        validation_loss = val_loss_accumulator * \
            val_loader.batch_size / len(val_loader.dataset)
        validation_accuracy = val_accuracy_accumulator * \
            val_loader.batch_size / len(val_loader.dataset)

        return validation_loss, validation_accuracy

    def run_test(self, data_loader, plot=False, save_path=None, pos_label=1,
                 transform_callable=None, **kwargs):
        """
        Will conduct the test suite to determine network strength. Using
        metrics.run_test

        Parameters:
            data_loader : DataLoader
                A DataLoader object to run the test with.
            save_path : string
                Folder to place images in.
            plot: bool
                Determine if graphs should be plotted in real time.
            pos_label: int
                The label that is positive in the binary case for macro
                calculations.
            transform_callable: callable
                A torch function. e.g. torch.round()
            kwargs: dict of keyworded parameters
                Values passed to transform callable (function parameters)

        Returns:
            results : dict

        """
        return self.metrics.run_test(
            network=self,
            data_loader=data_loader,
            save_path=save_path,
            plot=plot,
            pos_label=pos_label,
            transform_callable=transform_callable,
            **kwargs
        )

    # TODO: need to update transform callable params to match that of
    # cross_validate
    def bootfold_p_estimate(self, data_loader, n_samples, k, epochs, 
                            index_to_iter, ls_feat_vals, retain_graph=None,
                            valid_interv=4, plot=False, save_path=None,
                            transform_outputs=False, transform_callable=None, p_output_path = None,
			    **kwargs):
        """
	Performs bootfold - estimation to identify whether training model provides statistically significant
	difference in predicting various values for a given feature when predicting outcome. 

	Parameters:
	    network : BaseNetwork
                Network descendant of BaseNetwork.
            data_loader : torch.utils.data.DataLoader
                The DataLoader object containing the totality of the data to use
                for k-fold cross validation.
            n_samples : int
                number of times to randomly sample w/ replacement the data_loader and perform boot_cv
            k : int
                The number of folds to split the training into.
            epochs : int
                The number of epochs to train the network per fold.
	    index_to_iter : string
		Index of feature within data_loader who's values will be iterated to assess difference
	    ls_feat_vals : list
		List of values for feature provided in feat_to_iter
            valid_interv : int
                Specifies after how many epochs validation should occur.
            plot : boolean
                Whether or not to plot all results in prompt and charts.
            save_path : str
                Where to save all figures and results.
            transform_outputs : boolean
                Not used in the multi-class case.
                If true, transform outputs using metrics.transform_outputs.
                If no transform_callable is provided then the defaults in
                metrics.transform_outputs will be used: class converstion for
                one-hot encoded, and identity for one-dimensional outputs.
                Multiple class multiple outputs are not yet supported.
            transform_callable: callable
                Not used in the multi-class case.
                Used to transform values if transform_outputs is true,
                otherwise defaults in metrics.transform_outputs will be used.
                An example could be np.round
	    p_output_path = str
		Output file to save p_value to
            kwargs: dict of keyworded parameters
                Values passed to transform callable (function parameters)

	Returns:
	    p_value : float
        """
        return self.metrics.bootfold_p_estimate(
            network=self,
            data_loader=data_loader,
            n_samples=n_samples,
            k=k,
            epochs=epochs,
            index_to_iter=index_to_iter,
            ls_feat_vals=ls_feat_vals,
            retain_graph=retain_graph,
            valid_interv=valid_interv,
            plot=plot,
            save_path=save_path,
            transform_outputs=transform_outputs,
            transform_callable=transform_callable,
            p_output_path=p_output_path,
            **kwargs)

    def cross_validate(self, data_loader, k, epochs,
                       average_results=True, retain_graph=None,
                       valid_interv=4, plot=False, save_path=None,
                       transform_callable=None, **kwargs):
        """
        Perform k-fold cross validation given a Network and DataLoader object.

        Parameters:
            network : BaseNetwork
                Network descendant of BaseNetwork.
            data_loader : torch.utils.data.DataLoader
                The DataLoader object containing the totality of the data to use
                for k-fold cross validation.
            k : int
                The number of folds to split the training into.
            epochs : int
                The number of epochs to train the network per fold.
            average_results : boolean
                Whether or not to return results from all folds or just an average.
            retain_graph : {None, boolean}
                Whether retain_graph will be true when .backwards is called.
            valid_interv : int
                Specifies after how many epochs validation should occur.
            plot : boolean
                Whether or not to plot all results in prompt and charts.
            save_path : str
                Where to save all figures and results.
            transform_callable: callable
                A torch function. e.g. torch.round()
            kwargs: dict of keyworded parameters
                Values passed to transform callable (function parameters)


        Returns:
            results : dict
                If average_results is on, return dict of floats.
                If average_results is off, return dict of float lists.

        """
        # TODO: deal with repeated default parameters
        return self.metrics.cross_validate(
            network=self,
            data_loader=data_loader,
            k=k,
            epochs=epochs,
            average_results=average_results,
            retain_graph=retain_graph,
            valid_interv=valid_interv,
            plot=plot,
            save_path=save_path,
            transform_callable=transform_callable,
            **kwargs)

    def forward(self, inputs, **kwargs):
        """
        Perform a forward pass through the modules.

        If the network is defined with `num_classes` then it contains a
        classification layer/network tail. The inputs will be passed
        through the networks and then through the classifier.
        If not, the input is passed through the network and
        returned without passing through a classification layer.

        Parameters:
            inputs : list(torch.Tensor)
                The inputs to pass throught the network.

        Returns:
            output : torch.Tensor

        """
        if not isinstance(inputs, list):
            inputs = [inputs]

        if self.input_networks:
            net_outs = []
            # Loop through all input networks and pass through respective
            # input data tensors to collect their outputs. Use the specified
            # merge_inputs functionality to combine all the outputs to create
            # the input for this network.
            for in_net, x in zip(self.input_networks.values(), inputs):
                net_outs.append(in_net(x))
            output = self._merge_input_network_outputs(net_outs)
        else:
            output = torch.cat(inputs, dim=1)

        output = set_tensor_device(output, device=self.device)
        # Return actionable error if input shapes don't match up.
        if output.shape[1:] != self.in_dim:
            raise ValueError(
                "Input data incorrect dimension shape for network: {}. "
                "Expecting shape {} but recieved shape {}".format(
                    self.name, self.in_dim, output.shape[1:]))
        return self.network(output)

    @torch.no_grad()
    def forward_pass(self, data_loader,
                     transform_callable=None, **kwargs):
        """
        Allow the user to pass data through the network.
        Parameters:
            data_loader : DataLoader
                DataLoader object to make the pass with.
            transform_callable: callable
                A torch function. e.g. torch.round()
            kwargs: dict of keyworded parameters
                Values passed to transform callable (function parameters)
        Returns:
            outputs : numpy.ndarray
                Numpy matrix with the output. Same shape as network out_dim.
        """
        self.eval()
        # prediction_shape used to aggregate network outputs
        # (e.g. with or without class conversion)
        # so far always a float.
        dtype = torch.float
        pred_collector = torch.tensor([], dtype=dtype, device=self.device)
        for data, _ in data_loader:
            # Get raw network output
            predictions = self(data)

            if self._final_transform:
                predictions = self._final_transform(predictions)

            if transform_callable:
                predictions = transform_callable(predictions)

            # Aggregate predictions
            pred_collector = torch.cat([pred_collector, predictions])

        pred_collector = pred_collector.cpu().detach().numpy()
        return pred_collector


    # TODO: could integrate map location in the future if needed
    # https://discuss.pytorch.org/t/on-a-cpu-device-how-to-load-checkpoint-saved-on-gpu-device/349
    def save_model(self, save_path=None):
        """
        Save the model (and it's input networks).

        Parameters:
            save_path : str
                The save directory (not a file)

        Returns:
            save_path : str
                The save path where you'll find the model directly.

        """
        if not save_path:
            save_path = r"saved_models/"

        if not save_path.endswith("/"):
            save_path = save_path + "/"

        save_path = save_path + "{}_{}/".format(
            self.name, datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        logger.info("No save path provided, saving to {}".format(save_path))
        # Recursively save the input networks as well.
        if self.input_networks:
            for in_net in self.input_networks.values():
                in_net.save_model(save_path)

        if not os.path.exists(save_path):
            os.makedirs(save_path)

        self.save_path = save_path

        # TODO: to improve: # object.__getstate__()
        model_file_path = save_path + "model.pkl"
        state_dict_file_path = save_path + "state_dict.pkl"
        pickle.dump(self, open(model_file_path, "wb"), 2)
        pickle.dump(self.state_dict, open(state_dict_file_path, "wb"), 2)
        return self.save_path

    @classmethod
    def load_model(cls, load_path, load_complete_model_stack=True):
        """
        Load the model from the given directory.

        Parameters:
            load_path : str
                The load directory (not a file)
            load_complete_model_stack : boolean
                Whether to load all parent networks as well. Not yet implemented.

        Returns:
            network : BaseNetwork
                A network object with all components intact.

        """
        # TODO: does this break windows?? no idea.
        if not load_path.endswith("/"):
            load_path = load_path + "/"

        model_file_path = load_path + "model.pkl"

        instance = pickle.load(open(model_file_path, 'rb'))

        return instance



"""Define tests for ensemble networks such as Snapshot ensembling."""
import pytest
import numpy as np
import torch
from vulcanai.models.cnn import ConvNet
from vulcanai.models.dnn import DenseNet
from vulcanai.models.ensemble import SnapshotNet
from torch.utils.data import TensorDataset, DataLoader

torch.manual_seed(1234)


class TestSnapshotNet:
    """Test SnapshotNet functionality."""

    @pytest.fixture
    def cnn_noclass(self):
        """Create intermediate conv module."""
        return ConvNet(
            name='Test_ConvNet_noclass',
            in_dim=(1, 28, 28),
            config={
                'conv_units': [
                    {
                        "in_channels": 1,
                        "out_channels": 16,
                        "kernel_size": (5, 5),
                        "stride": 2
                    },
                    {
                        "in_channels": 16,
                        "out_channels": 1,
                        "kernel_size": (5, 5),
                        "stride": 1,
                        "padding": 2
                    }]
            },
        )

    @pytest.fixture
    def dnn_class(self, cnn_noclass):
        """Create dnn module prediction leaf node."""
        return DenseNet(
            name='Test_DenseNet_class',
            input_networks=cnn_noclass,
            config={
                'dense_units': [100, 50],
                'initializer': None,
                'bias_init': None,
                'norm': None,
                'dropout': 0.5,  # Single value or List
            },
            num_classes=3
        )

    @pytest.fixture
    def dnn_class_two(self, cnn_noclass):
        """Create dnn module prediction leaf node."""
        return DenseNet(
            name='Test_DenseNet_class',
            input_networks=cnn_noclass,
            optim_spec={'name': 'Adam', 'lr': 0.001},
            config={
                'dense_units': [100, 50],
                'initializer': None,
                'bias_init': None,
                'norm': None,
                'dropout': 0.5,  # Single value or List
            },
            num_classes=3
        )

    def test_snapshot_structure(self, cnn_noclass, dnn_class):
        """Confirm Snapshot structure is generated properly."""
        test_input = torch.randint(0, 10, size=[3, *cnn_noclass.in_dim]).float()
        test_target = torch.LongTensor([0, 2, 1])
        test_dataloader = DataLoader(TensorDataset(test_input, test_target))
        test_snap = SnapshotNet(
            name='test_snap',
            template_network=dnn_class,
            n_snapshots=3
        )
        test_snap.fit(
            train_loader=test_dataloader,
            val_loader=test_dataloader,
            epochs=3,
            plot=False
        )
        assert isinstance(
            test_snap.template_network.lr_scheduler,
            torch.optim.lr_scheduler.CosineAnnealingLR)
        # Check correct number of generated snapshots
        assert len(test_snap.network) == 3
        # Check snapshots are not identical
        assert test_snap.network[0] is not \
            test_snap.network[1] is not \
            test_snap.network[2]
        output = test_snap.forward_pass(
            data_loader=test_dataloader,
            transform_outputs=False)
        assert output.shape == (3, test_snap._num_classes)
        assert np.any(~np.isnan(output))

    def test_snapshot_lr(self, cnn_noclass, dnn_class_two):
        """Confirm Snapshot structure is generated properly."""
        test_input = torch.randint(0, 10, size=[3, *cnn_noclass.in_dim]).float()
        test_target = torch.LongTensor([0, 2, 1])
        test_dataloader = DataLoader(TensorDataset(test_input, test_target))
        test_snap = SnapshotNet(
            name='test_snap',
            template_network=dnn_class_two,
            n_snapshots=3,
        )
        test_snap.fit(
            train_loader=test_dataloader,
            val_loader=test_dataloader,
            epochs=15,
            plot=False
        )
        assert test_snap.template_network.lr_scheduler.get_lr()[0] < 0.001



# coding=utf-8
""" Defines test cases for tabular dataset """
import pytest
from vulcanai.datasets import TabularDataset
import os
import pandas as pd


# noinspection PyMissingOrEmptyDocstring
class TestTabularDataset:

    @pytest.fixture
    def split_data(self):
        pass

    @pytest.fixture
    def my_test_dataset(self):
        """Create a dataset by importing from the test csv"""
        fpath = str(os.path.dirname(__file__)) + \
            "/test_data/birthweight_reduced.csv"
        return TabularDataset(
            data=fpath,
            label_column="id",
            na_values='Nan'
        )

    @pytest.fixture
    def my_test_dataset_two(self):
        """Create a second dataset by importing from the test csv"""
        fpath = str(os.path.dirname(__file__)) + \
            "/test_data/birthweight_reduced2.csv"
        return fpath

    @pytest.fixture
    def my_merged_test_dataset(self):
        fpath = str(os.path.dirname(__file__)) + \
            "/test_data/birthweight_reduced_merged.csv"
        return pd.read_csv(fpath)

    def test_merge_dataframe(self, my_test_dataset, my_test_dataset_two,
                             my_merged_test_dataset):
        dct_df = {'df_test_two': my_test_dataset_two}
        my_test_dataset.merge_dataframe(merge_on_columns=['id'],
                                        na_values='Nan', **dct_df)
        pd.testing.assert_frame_equal(my_test_dataset.df,
                                      my_merged_test_dataset,
                                      check_dtype=False)

    def test_single_dataset_length(self, my_test_dataset):
        assert len(my_test_dataset) == 42
        assert "id" in my_test_dataset.label_column

    def test_save_dataframe(self, my_test_dataset):
        fpath = str(os.path.dirname(__file__)) + "/test_data/test_save.csv"
        my_test_dataset.save_dataframe(fpath)
        assert os.path.isfile(fpath)
        os.remove(fpath)

    def test_list_columns(self, my_test_dataset):
        column_list = ["id", "headcirumference", "length", "Birthweight",
                       "LowBirthWeight", "Gestation", "smoker", "motherage",
                       "mnocig", "mheight", "mppwt", "fage", "fedyrs"]
        assert set(my_test_dataset.list_all_features()) == set(column_list)

    def test_replace_value_in_column(self, my_test_dataset):
        before = my_test_dataset.list_all_column_values("motherage")
        my_test_dataset.replace_value_in_column("motherage", 24, 25)
        after = my_test_dataset.list_all_column_values("motherage")
        assert (set(before) - set(after)) == {24}

    def test_delete_columns(self, my_test_dataset):
        before = my_test_dataset.list_all_features()
        my_test_dataset.delete_column("motherage")
        after = my_test_dataset.list_all_features()
        assert (set(before) - set(after)) == {"motherage"}

    def test_create_label_encoding(self, my_test_dataset):
        my_test_dataset.create_label_encoding("LowBirthWeight",
                                              {"Low": 0, "Normal": 1})
        assert set(my_test_dataset.list_all_column_values("LowBirthWeight")) \
            == {0, 1}

    def test_create_one_hot_encoding(self, my_test_dataset):
        my_test_dataset.create_one_hot_encoding("LowBirthWeight")
        assert "LowBirthWeight@Low" in my_test_dataset.list_all_features()

    def test_reverse_create_all_one_hot_encodings(self, my_test_dataset):
        my_test_dataset.create_one_hot_encoding("LowBirthWeight")
        my_test_dataset.reverse_create_one_hot_encoding(column_name=None)
        assert "LowBirthWeight@Low" not in my_test_dataset.list_all_features()

    def test_identify_null(self, my_test_dataset):
        num_threshold = 0.2
        res = my_test_dataset.identify_null(num_threshold)
        assert {'fedyrs'} == set(res)

    def test_identify_unique(self, my_test_dataset):
        res = my_test_dataset.identify_unique(5)
        assert set(res) == {'headcirumference', 'smoker', 'fedyrs',
                            'LowBirthWeight'}

    def test_identify_unbalanced_columns(self, my_test_dataset):
        res = my_test_dataset.identify_unbalanced_columns(0.5)
        assert set(res) == {'headcirumference', 'smoker', 'mnocig',
                            'LowBirthWeight'}

    def test_identify_highly_correlated(self, my_test_dataset):
        res = my_test_dataset.identify_highly_correlated(0.2)
        assert (('fage', 'motherage'), 0.8065844173531495) in res

    def test_identify_low_variance(self, my_test_dataset):
        res = my_test_dataset.identify_low_variance(0.05)
        assert 'Gestation' in res

    def test_split_length_correct(self, my_test_dataset):
        res = my_test_dataset.split([0.1, 0.2, 0.7])
        assert len(res) == 3
        assert isinstance(res[1], TabularDataset)
        res = my_test_dataset.split([0.1, 0.9])
        assert len(res) == 2
        assert isinstance(res[1], TabularDataset)

    def test_split_length_correct_stratified(self, my_test_dataset):
        res = my_test_dataset.split([0.1, 0.2, 0.7], stratified=True,
                                    stratum_column="motherage")
        assert len(res) == 3
        assert isinstance(res[1], TabularDataset)
        res = my_test_dataset.split([0.1, 0.9], stratified=True,
                                    stratum_column="motherage")
        assert len(res) == 2
        assert isinstance(res[1], TabularDataset)

    def test_stratified_values(self, my_test_dataset):
        res = my_test_dataset.split([0.1, 0.2, 0.7], stratified=True,
                                    stratum_column="motherage")

        assert list(res[0].df["motherage"].values) == [20]
        assert list(res[1].df["motherage"].values) == [19, 20, 21, 24, 27, 31]
        assert list(res[2].df["motherage"].values) == [18, 19, 19, 20, 20, 20,
                                                       20, 20, 21, 21, 22, 22,
                                                       23, 23, 24, 24, 24, 26,
                                                       26, 27, 27, 27, 28, 28,
                                                       29, 29, 30, 30, 31, 31,
                                                       32, 35, 37, 37, 41]



"""Test device switching for networks."""
import pytest

import torch
from torch.utils.data import DataLoader, Subset

from vulcanai.models import ConvNet, DenseNet
from vulcanai.models.utils import master_device_setter

TEST_CUDA = torch.cuda.is_available()
TEST_MULTIGPU = TEST_CUDA and torch.cuda.device_count() >= 2
DEVICE_COUNT = 0

if TEST_CUDA:
    DEVICE_COUNT = torch.cuda.device_count()


class TestDevice:
    """Test multi-input GPU device switching."""

    @pytest.mark.skipif(not TEST_CUDA, reason="No CUDA"
                        " supported devices available")
    def test_master_net_device_set_to_cuda(self, multi_input_cnn):
        """Test if the network as whole gets switched to cuda."""
        assert hasattr(multi_input_cnn, 'device')
        master_device_setter(multi_input_cnn, 'cuda:0')
        assert multi_input_cnn.device == torch.device(type='cuda', index=0)
        assert multi_input_cnn.input_networks['conv3D_net']\
            .device == torch.device(type='cuda', index=0)
        assert multi_input_cnn.input_networks['multi_input_dnn']\
            .device == torch.device(type='cuda', index=0)
        assert multi_input_cnn.input_networks['multi_input_dnn'].\
            input_networks['conv1D_net'].\
            device == torch.device(type='cuda', index=0)
        assert multi_input_cnn.input_networks['multi_input_dnn'].\
            input_networks['conv2D_net'].\
            device == torch.device(type='cuda', index=0)

    @pytest.mark.skipif(not TEST_CUDA, reason="No CUDA"
                        " supported devices available")
    def test_fail_mixed_devices(self, multi_input_cnn, conv3D_net,
                                multi_input_dnn, conv1D_net,
                                multi_input_dnn_data,
                                multi_input_cnn_data):
        """Test training throws ValueError when network has mixed devices."""
        assert hasattr(conv1D_net, 'device')
        assert hasattr(conv3D_net, 'device')
        assert hasattr(multi_input_dnn, 'device')
        assert hasattr(multi_input_cnn, 'device')

        master_device_setter(multi_input_cnn, 'cuda:0')
        assert conv3D_net == multi_input_cnn.input_networks['conv3D_net']
        assert multi_input_dnn == multi_input_cnn.input_networks['multi_input_dnn']

        data_len = len(multi_input_cnn_data)
        train_loader = DataLoader(
            Subset(multi_input_cnn_data, range(data_len//2)))
        valid_loader = DataLoader(
            Subset(multi_input_cnn_data, range(data_len//2, data_len)))

        multi_input_cnn.fit(
            train_loader=train_loader,
            val_loader=valid_loader,
            epochs=1,
            plot=False)

        with pytest.raises(ValueError) as e_info:
            multi_input_cnn.input_networks['conv3D_net'].device = 'cpu'
            multi_input_cnn.fit(
                train_loader=train_loader,
                val_loader=valid_loader,
                epochs=1,
                plot=False)

        assert str(e_info.value).endswith("{'conv3D_net': device(type='cpu')}")



# coding=utf-8
"""Defines the ConvNet class."""
import torch
import torch.nn as nn
import numpy as np

from .basenetwork import BaseNetwork
from .layers import DenseUnit, ConvUnit, FlattenUnit
from .utils import pad

import logging
from inspect import getfullargspec
from math import ceil

from collections import OrderedDict

logger = logging.getLogger(__name__)


# TODO: use setters to enforce types/formats/values!
# TODO: make this a base class?
# TODO: add additional constraints in the future
class ConvNetConfig:
    """Define the necessary configuration for a ConvNet."""

    def __init__(self, raw_config):
        """
        Take in user config dict and clean it up.

        Cleaned units is stored in self.units

        Parameters:
            raw_config : dict of dict
                User specified dict

        """
        if 'conv_units' not in raw_config:
            raise KeyError("conv_units must be specified")

        # Confirm all passed units conform to Unit required arguments
        conv_unit_arg_spec = getfullargspec(ConvUnit)
        conv_unit_arg_spec.args.remove('self')
        # Deal with dim inference when cleaning unit
        conv_unit_arg_spec.args.remove('conv_dim')

        # Find the index for where the defaulted values begin
        default_arg_start_index = len(conv_unit_arg_spec.args) - \
            len(conv_unit_arg_spec.defaults)
        self.required_args = conv_unit_arg_spec.args[:default_arg_start_index]
        self.units = []
        for u in raw_config['conv_units']:
            unit = self._clean_unit(raw_unit=u)
            self.units.append(unit)

    def _clean_unit(self, raw_unit):
        """
        Use this to catch mistakes in each user-specified unit.

        Infer dimension of Conv using the kernel shape.

        Parameters:
            raw_unit : dict

        Returns:
            unit : dict
                Cleaned unit config.

        """
        unit = raw_unit
        for key in self.required_args:
            if key not in unit.keys():
                raise ValueError(
                    "{} needs to be specified in your config.".format(key))
        if not isinstance(unit['kernel_size'], tuple):
            if isinstance(unit['kernel_size'], int):
                unit['kernel_size'] = (unit['kernel_size'],)
            unit['kernel_size'] = tuple(unit['kernel_size'])
        unit['conv_dim'] = len(unit['kernel_size'])
        return unit


class ConvNet(BaseNetwork):
    """
    Subclass of BaseNetwork defining a ConvNet.

    Parameters:
        name : str
            The name of the network. Used when saving the file.
        config : dict
            The configuration of the network module, as a dict.
        in_dim : tuple
            The input dimensions of the network. Not required to specify when the
            network has input_networks.
        save_path : str
            The name of the file to which you would like to save this network.
        input_networks : list of BaseNetwork
            A network object provided as input.
        num_classes : int or None
            The number of classes to predict.
        activation : torch.nn.Module
            The desired activation function for use in the network.
        pred_activation : torch.nn.Module
            The desired activation function for use in the prediction layer.
        optim_spec : dict
            A dictionary of parameters for the desired optimizer.
        lr_scheduler : torch.optim.lr_scheduler
            A callable torch.optim.lr_scheduler
        early_stopping : str or None
            So far just 'best_validation_error' is implemented.
        early_stopping_patience: integer
            Number of validation iterations of decreasing loss
            (note -not necessarily every epoch!
            before early stopping is applied.
        early_stopping_metric: string
            Either "loss" or "accuracy" are implemented.
        criter_spec : dict
            criterion specification with name and all its parameters.

    Returns:
        network : ConvNet
            A network of type BaseNetwork.

    """

    def __init__(self, name, config, in_dim=None, save_path=None,
                 input_networks=None, num_classes=None,
                 activation=nn.ReLU(), pred_activation=None,
                 optim_spec={'name': 'Adam', 'lr': 0.001},
                 lr_scheduler=None, early_stopping=None,
                 early_stopping_patience=None,
                 early_stopping_metric="accuracy",
                 criter_spec=nn.CrossEntropyLoss(),
                 device="cuda:0"):
        """Define the ConvNet object."""
        super(ConvNet, self).__init__(
            name, ConvNetConfig(config), in_dim, save_path, input_networks,
            num_classes, activation, pred_activation, optim_spec,
            lr_scheduler, early_stopping, early_stopping_patience,
            early_stopping_metric,
            criter_spec, device)

    def _create_network(self, **kwargs):
        """
        Build the layers of the network into a nn.Sequential object.

        Parameters:
            conv_hid_layers : ConvNetConfig.units (list of dict)
                The hidden layers specification
            activation : torch.nn.Module
                the non-linear activation to apply to each layer

        Returns:
            output : torch.nn.Sequential
                the conv network as a nn.Sequential object

        """
        conv_hid_layers = self._config.units

        if self.input_networks:
            self.in_dim = self._get_in_dim()

        conv_hid_layers[0]['in_channels'] = self.in_dim[0]
        conv_layers = OrderedDict()
        for idx, conv_layer_config in enumerate(conv_hid_layers):
            conv_layer_config['activation'] = kwargs['activation']
            layer_name = 'conv_{}'.format(idx)
            conv_layers[layer_name] = ConvUnit(**conv_layer_config)
        self.network = nn.Sequential(conv_layers)

        if self._num_classes:
            self.network.add_module(
                'flatten', FlattenUnit())
            self.network.add_module(
                'classify', DenseUnit(
                    in_features=self._get_out_dim()[0],
                    out_features=self._num_classes,
                    activation=kwargs['pred_activation']))

    def _merge_input_network_outputs(self, tensors):
        """Calculate converged in_dim for the MultiInput ConvNet."""
        reshaped_tensors = []
        # Determine what shape to cast to without losing any information.
        max_conv_tensor_size = self._get_max_incoming_spatial_dims()
        for t in tensors:
            if t.dim() == 2:
                # Cast Linear output to largest Conv output shape
                t = self._cast_linear_to_shape(
                    tensor=t,
                    cast_shape=max_conv_tensor_size)
            elif t.dim() > 2:
                # Cast Conv output to largest Conv output shape
                t = self._cast_conv_to_shape(
                    tensor=t,
                    cast_shape=max_conv_tensor_size)
            reshaped_tensors.append(t)
        return torch.cat(reshaped_tensors, dim=1)

    def _get_max_incoming_spatial_dims(self):
        """Return the max spatial dimensions of the input networks."""
        # Ignoring the channels
        spatial_inputs = []
        for in_net in self.input_networks.values():
            if isinstance(in_net, ConvNet):
                spatial_inputs.append(list(in_net.out_dim[1:]))
        max_spatial_dim = len(max(spatial_inputs, key=len))

        # Fill with zeros in missing dim to compare
        # max size later for each dim.
        for in_spatial_dim in spatial_inputs:
            while(len(in_spatial_dim) < max_spatial_dim):
                in_spatial_dim.insert(0, 0)

        # All spatial dimensions
        # Take the max size in each dimension.
        max_conv_tensor_size = np.array(spatial_inputs).transpose().max(axis=1)
        return np.array(max_conv_tensor_size)

    def _cast_linear_to_shape(self, tensor, cast_shape):
        """
        Convert Linear outputs into Conv outputs.

        Parameters:
            tensor : torch.Tensor
                The Linear tensor to reshape of shape [out_features]
            cast_shape : numpy.ndarray
                The Conv shape to cast linear to of shape
                [batch, num_channels, *spatial_dimensions].

        Returns:
            tensor : torch.Tensor
                Tensor of shape [batch, num_channels, *spatial_dimensions]

        """
        # Equivalent to calculating tensor.numel() in pytorch.
        sequence_length = cast_shape.prod()
        # How many channels to spread the information into
        # Ignore batch from linear
        n_channels = ceil(tensor[-1].numel() / sequence_length)
        # How much pad to add to either sides to reshape the linear tensor
        # into cast_shape spatial dimensions.
        pad_shape = sequence_length * n_channels
        tensor = pad(tensor=tensor, target_shape=[pad_shape])
        return tensor.view(-1, n_channels, *cast_shape)

    def _cast_conv_to_shape(self, tensor, cast_shape):
        """
        Convert Conv outputs into Conv outputs.

        Parameters:
            tensor : torch.Tensor
                The Conv tensor to reshape of shape
                [batch, num_channels, *spatial_dimensions]
            cast_shape : numpy.ndarray
                The Conv shape to cast incoming Conv to shape
                [batch, num_channels, *spatial_dimensions].

        Returns:
            tensor : torch.Tensor
                Tensor of shape [batch, num_channels, *spatial_dimensions]

        """
        # Extract only the spatial dimensions by ignoring the batch and channel
        spatial_dim_idx_start = 2
        if len(tensor.shape[spatial_dim_idx_start:]) < len(cast_shape):
            # TODO: https://github.com/pytorch/pytorch/issues/9410
            # Ignore batch for incoming tensor
            # For each missing dim, add dims until it
            # is equivalient to the max dim
            n_unsqueezes = len(cast_shape) - \
                len(tensor.shape[spatial_dim_idx_start:])
            for _ in range(n_unsqueezes):
                tensor = tensor.unsqueeze(dim=spatial_dim_idx_start)
        return pad(tensor=tensor, target_shape=cast_shape)

    def __str__(self):
        """Specify how to print network."""
        if self.optim is not None:
            return super(ConvNet, self).__str__() + '\noptim: {}'.format(self.optim)
        else:
            return super(ConvNet, self).__str__()



"""Contains all visualization methods."""
import os

import numpy as np

from math import sqrt, ceil, floor
import pickle
from datetime import datetime
from .utils import GuidedBackprop, get_notable_indices

from sklearn.manifold import TSNE
from sklearn.decomposition import PCA

import matplotlib
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.pyplot as plt
import seaborn as sns

import itertools
import logging
logger = logging.getLogger(__name__)

DISPLAY_AVAILABLE = True if os.environ.get("DISPLAY") else False


def save_visualization(plot, path=None):
    """
    Save plot at designated path.

    Parameters:
        plot : matplotlib
            Matplotlib variable with savefig ability.
        path : string
            String that designates the path to save the given figure to.

    Returns:
        None

    """
    plot.savefig(path)
    logger.info("Saved visualization at %s", path)


def get_save_path(path, vis_type):
    """Return a save_path string."""
    path = "{}{}_{date:%Y-%m-%d_%H-%M-%S}.png".format(
        path, vis_type, date=datetime.now())
    return path


def display_record(record=None, save_path=None, interactive=True):
    """
    Display the training curve for a network training session.

    Parameters:
        record : dict
            the network record dictionary for dynamic graphs during training.
        save_path : String
            String that designates the path to save figure to be produced.
            Save_path must be a proper path that ends with a filename with an
            image filetype.
        interactive : boolean
            To display during training or afterwards.

    Returns:
        None

    """
    title = 'Training curve'
    if record is None or not isinstance(record, dict):
        raise ValueError('No record exists and cannot be displayed.')

    plt.subplot(1, 2, 1)
    plt.title("{}: Error".format(title))
    train_error, = plt.plot(
        record['epoch'],
        record['train_error'],
        '-mo',
        label='Train Error'
    )
    # val_error = \
    # [i if ~np.isnan(i) else None for i in record['validation_error']]
    validation_error, = plt.plot(
        record['epoch'],
        record['validation_error'],
        '-ro',
        label='Validation Error'
    )
    plt.xlabel("Epoch")
    plt.ylabel("Cross entropy error")
    plt.legend(handles=[train_error,
                        validation_error],
               loc=0)

    plt.subplot(1, 2, 2)
    plt.title("{}: Accuracy".format(title))
    train_accuracy, = plt.plot(
        record['epoch'],
        record['train_accuracy'],
        '-go',
        label='Train Accuracy'
    )
    validation_accuracy, = plt.plot(
        record['epoch'],
        record['validation_accuracy'],
        '-bo',
        label='Validation Accuracy'
    )
    plt.xlabel("Epoch")
    plt.ylabel("Accuracy")
    plt.ylim(0, 1)

    plt.legend(handles=[train_accuracy,
                        validation_accuracy],
               loc=0)

    if save_path:
        save_visualization(plt, save_path)

    if not DISPLAY_AVAILABLE and save_path is None:
        raise RuntimeError(
            "No display environment found. "
            "Display environment needed to plot, "
            "or set save_path=path/to/dir")

    elif interactive is True:
        plt.draw()
        plt.pause(1e-17)


def display_pca(input_data, targets, label_map=None, save_path=None):
    """
    Calculate pca reduction and plot it.

    Parameters:
        input_data : numpy.dnarray
            Input data to reduce in dimensions.
        targets : numpy.ndarray
            size (batch, labels) for samples.
        label_map : dict
            labelled {str(int), string} key, value pairs.
        save_path : String
            String that designates the path to save figure to be produced.

    """
    pca = PCA(n_components=2, random_state=0)
    x_transform = pca.fit_transform(input_data)
    _plot_reduction(
        x_transform,
        targets,
        label_map=label_map,
        title='PCA Visualization',
        save_path=save_path)


def display_tsne(input_data, targets, label_map=None, save_path=None):
    """
    t-distributed Stochastic Neighbor Embedding (t-SNE) visualization [1].

    [1]: Maaten, L., Hinton, G. (2008). Visualizing Data using t-SNE.
            JMLR 9(Nov):2579--2605.

    Parameters:
        input_data : numpy.dnarray
            Input data to reduce in dimensions.
        targets : numpy.ndarray
            size (batch, labels) for samples.
        label_map : dict
            labelled {str(int), string} key, value pairs.
        save_path : String
            String that designates the path to save figure to be produced.

    """
    tsne = TSNE(n_components=2, random_state=0)
    x_transform = tsne.fit_transform(input_data)
    _plot_reduction(
        x_transform,
        targets,
        label_map=label_map,
        title='t-SNE Visualization',
        save_path=save_path)


def _plot_reduction(x_transform, targets, label_map, title, save_path=None):
    """Once PCA and t-SNE has been calculated, this is used to plot."""
    y_unique = np.unique(targets)
    if label_map is None:
        label_map = {str(i): str(i) for i in y_unique}
    elif not isinstance(label_map, dict):
        raise ValueError('label_map most be a dict of a key'
                         ' mapping to its true label')
    colours = np.array(sns.color_palette("hls", len(y_unique)))
    plt.figure()
    for index, cl in enumerate(y_unique):
        plt.scatter(x=x_transform[targets == cl, 0],
                    y=x_transform[targets == cl, 1],
                    s=100,
                    c=colours[index],
                    alpha=0.5,
                    marker='o',
                    edgecolors='none',
                    label=label_map[str(cl)])
    plt.xlabel('X')
    plt.ylabel('Y')
    plt.legend(loc='upper right')
    plt.title(title)

    if save_path:
        save_path = get_save_path(save_path, vis_type=title)
        save_visualization(plt, save_path)

    if not DISPLAY_AVAILABLE and save_path is None:
        raise RuntimeError(
            "No display environment found. "
            "Display environment needed to plot, "
            "or set save_path=path/to/dir")
    else:
        plt.show(False)


def display_confusion_matrix(cm, class_list=None, save_path=None):
    """
    Print and plot the confusion matrix.

    inspired from: https://github.com/zaidalyafeai/Machine-Learning

    Parameters:
        cm : numpy.ndarray
            2D confustion_matrix obtained using utils.get_confusion_matrix
        class_list : list
            Actual class labels (e.g.: MNIST - [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        save_path : String
            String that designates the path to save figure to be produced.

    """
    if class_list is None:
        class_list = list(range(cm.shape[0]))
    if not isinstance(class_list, list):
        raise ValueError("class_list must be of type list.")
    plt.figure()
    ax = plt.gca()
    im = ax.imshow(cm, interpolation='nearest', cmap='Blues', origin='lower')

    plt.title('Confusion matrix')
    tick_marks = np.arange(len(class_list))
    plt.xticks(tick_marks, class_list, rotation=45)
    plt.yticks(tick_marks, class_list)
    # Plot number overlay
    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, format(cm[i, j], 'd'),
                 horizontalalignment="center",
                 color="white" if cm[i, j] > thresh else "black")
    # Plot labels
    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="4%", pad=0.03)
    plt.colorbar(im, cax=cax)

    if save_path:
        save_path = get_save_path(save_path, vis_type='confusion_matrix')
        save_visualization(plt, save_path)

    if not DISPLAY_AVAILABLE and save_path is None:
        raise RuntimeError(
            "No display environment found. "
            "Display environment needed to plot, "
            "or set save_path=path/to/dir")
    else:
        plt.show(False)


def compute_saliency_map(network, input_data, targets):
    """
    Return the saliency map using the guided backpropagation method [1].

    [1]: Springgenberg, J.T., Dosovitskiy, A., Brox, T., Riedmiller, M. (2015).
         Striving for Simplicity: The All Convolutional Net. ICLR 2015
         (https://arxiv.org/pdf/1412.6806.pdf)

    Parameters:
        network : BaseNetwork
            A network to get saliency maps on.
        input_data : numpy.ndarray
            Input array of shape (batch, channel, width, height) or
            (batch, channel, width, height, depth).
        targets : numpy.ndarray
            1D array with class targets of size [batch].

    Returns:
        saliency_map : list of numpy.ndarray
            Top layer gradients of the same shape as input data.

    """
    guided_backprop = GuidedBackprop(network)
    saliency_map = guided_backprop.generate_gradients(input_data, targets)
    return saliency_map


def display_saliency_overlay(image, saliency_map, shape=(28, 28), save_path=None):
    """
    Plot overlay saliency map over image.

    Parameters:
        image : numpy.ndarray
            (1D, 2D, 3D) for single image or linear output.
        saliency_map: numpy.ndarray
            (1D, 2D, 3D) for single image or linear output.
        shape : tuple, list
            The dimensions of the image. Defaults to mnist.
        save_path : String
            String that designates the path to save figure to be produced.

    """
    # Handle different colour channels and shapes for image input
    if len(image.shape) == 3:
        if image.shape[0] == 1:
            # For 1 colour channel, remove it
            image = image[0]
        elif image.shape[0] == 3 or image.shape[0] == 4:
            # For 3 or 4 colour channels, move to end for plotting
            image = np.moveaxis(image, 0, -1)
        else:
            raise ValueError("Invalid number of colour channels in input.")
    elif len(image.shape) == 1:
        image = np.reshape(image, shape)

    # Handle different colour channels and shapes for saliency map
    if len(saliency_map.shape) == 3:
        if saliency_map.shape[0] == 1:
            # For 1 colour channel, remove it
            saliency_map = saliency_map[0]
        elif saliency_map.shape[0] == 3 or saliency_map.shape[0] == 4:
            # For 3 or 4 colour channels, move to end for plotting
            saliency_map = np.moveaxis(saliency_map, 0, -1)
        else:
            raise ValueError("Invalid number of channels in saliency map.")
    elif len(saliency_map.shape) == 1:
        saliency_map = np.reshape(saliency_map, shape)

    fig = plt.figure()
    fig.suptitle("Saliency Map")
    # Plot original image
    fig.add_subplot(1, 2, 1)
    plt.imshow(image, cmap='gray')
    # Plot original with saliency overlay
    fig.add_subplot(1, 2, 2)
    plt.imshow(image, cmap='binary')

    ax = plt.gca()
    im = ax.imshow(saliency_map, cmap='Blues', alpha=0.7)
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="4%", pad=0.03)
    plt.colorbar(im, cax=cax, format='%.0e')

    if save_path:
        save_path = get_save_path(save_path, vis_type='saliency_map')
        save_visualization(plt, save_path)

    if not DISPLAY_AVAILABLE and save_path is None:
        raise RuntimeError(
            "No display environment found. "
            "Display environment needed to plot, "
            "or set save_path=path/to/dir")
    else:
        plt.show(False)


def display_receptive_fields(network, top_k=5, save_path=None):
    """
    Display receptive fields of layers from a network [1].

    [1]: Luo, W., Li, Y., Urtason, R., Zemel, R. (2016).
         Understanding the Effective Receptive Field in Deep
         Convolutional Neural Networks. Advances in Neural Information
         Processing Systems, 29 (NIPS 2016)

    Parameters:
        network : BaseNetwork
            Network to get receptive fields of.
        top_k : int
            To return the most and least k important features from field
        save_path : String
            String that designates the path to save figure to be produced.

    Returns:
        k_features: dict
            A dict of the top k and bottom k important features.

    """
    if type(network).__name__ == "ConvNet":
        raise NotImplementedError
    elif '_input_network' in network._modules:
        if type(network._modules['_input_network']).__name__ == "ConvNet":
            raise NotImplementedError

    feature_importance = {}
    fig = plt.figure()
    fig.suptitle("Feature importance")
    num_layers = len(network._modules['network'])
    for i, layer in enumerate(network._modules['network']):
        raw_field = layer._kernel._parameters['weight'].cpu().detach().numpy()
        field = np.average(raw_field, axis=0)  # average all outgoing
        field_shape = [
            floor(sqrt(field.shape[0])),
            ceil(sqrt(field.shape[0]))
            ]
        fig.add_subplot(
            floor(sqrt(num_layers)),
            ceil(sqrt(num_layers)),
            i + 1)
        field = abs(field)
        feats = get_notable_indices(field, top_k=top_k)
        unit_type = type(layer).__name__
        layer_name = '{}_{}'.format(unit_type, i)
        feature_importance.update({layer_name: feats})
        plt.title(layer_name)
        plt.imshow(np.resize(field, field_shape), cmap='Blues')
        plt.colorbar()

    if save_path:
        save_path = get_save_path(save_path, vis_type='feature_importance')
        save_visualization(plt, save_path)

    if not DISPLAY_AVAILABLE and save_path is None:
        raise RuntimeError(
            "No display environment found. "
            "Display environment needed to plot, "
            "or set save_path=path/to/dir")
    else:
        plt.show(False)

    return feature_importance



"""Test BaseNetwork functionality."""
import pytest
from vulcanai.models.basenetwork import BaseNetwork

import torch


class TestBaseNetwork:
    """Define BaseNetwork test class."""

    @pytest.fixture
    def basenet(self):
        """Create a test BaseNetwork."""
        return BaseNetwork(
            name='Test_BaseNet',
            in_dim=(None, 10),
            config={}
        )

    def test_init(self, basenet):
        """Initialization Test of a BaseNetwork object."""
        assert isinstance(basenet, BaseNetwork)
        assert isinstance(basenet, torch.nn.Module)
        assert hasattr(basenet, 'network')
        assert hasattr(basenet, 'in_dim')
        assert hasattr(basenet, 'record')

    def test_name(self, basenet):
        """Test changing names."""
        basenet.name = 'New_Name'
        assert basenet.name is 'New_Name'

    def test_learning_rate(self, basenet):
        """Test learning rate change."""
        basenet.learning_rate = 0.1
        assert basenet.learning_rate is 0.1

    def test_default_criter_spec(self, basenet):
        """Test default value behaviour for criter spec"""
        assert isinstance(basenet.criter_spec, torch.nn.CrossEntropyLoss)
        assert isinstance(basenet._final_transform, torch.nn.Softmax)



"""Contains all visualization utilities."""
import numpy as np
import torch
from torch.nn import ReLU, SELU, ModuleList
import logging
import collections

logger = logging.getLogger(__name__)
# from ..models.basenetwork import BaseNetwork


def get_notable_indices(feature_importances, top_k=5):
    """
    Return dict of top k and bottom k features useful from matrix.

    Parameters:
        feature_importance : numpy.ndarray
            1D numpy array to extract the top of bottom indices.
        top_k : int
            How many features from top and bottom to extract.
            Defaults to 5.

    Returns:
        notable_indices : dict
            Indices of the top most important features.
            Indices of the bottom mos unimportant features.

    """
    important_features = feature_importances.argsort()[-top_k:][::-1]
    unimportant_features = feature_importances.argsort()[:-1][:top_k]
    return {'important_indices': important_features,
            'unimportant_indices': unimportant_features}


class GuidedBackprop(object):
    """
    Generate gradients with guided back propagation w.r.t given input.

    Modified from https://github.com/utkuozbulak/pytorch-cnn-visualizations
    Insert backward hooks for activations to propagate positive gradients.

    Parameters:
        network : BaseNetwork
            Network to conduct guided backprop on.

    Returns:
        gradients : list of numpy.ndarray
            Gradients of top most layer w.r.t the input sample.

    """

    def __init__(self, network):
        """Set up hooks for activations and gradient retrieval."""
        if network.__class__.__bases__[0].__name__ != "BaseNetwork":
            raise ValueError("Network type must be a subclass of BaseNetwork")
        self.network = network
        self.gradients = []
        self.hooks = []
        # Put model in evaluation mode
        self.network.eval()
        self._crop_negative_gradients()

    def _crop_negative_gradients(self):
        """Update relu/selu activations to return positive gradients."""
        def activation_hook_function(module, grad_in, grad_out):
            """If there is a negative gradient, changes it to zero."""
            if isinstance(module, ReLU) or isinstance(module, SELU):
                return (torch.clamp(grad_in[0], min=0.0),)
            else:
                raise NotImplementedError("Only ReLU and SELU supported.")

        def hook_all_networks(network):
            self.hooks.append(
                    network.network[0]._activation.
                    register_backward_hook(activation_hook_function))
            logging.info("Cropping gradients in {}.".format(network.name))
            if network.input_networks:
                for in_net in network.input_networks.values():
                    hook_all_networks(in_net)
        # For Snapshot Networks
        if isinstance(self.network.network, ModuleList):
            for net in self.network.network:
                hook_all_networks(net)
        else:
            hook_all_networks(self.network)

    def _remove_hooks(self):
        """Remove all previously placed hooks from model."""
        for h in self.hooks:
            h.remove()

    def generate_gradients(self, input_data, targets):
        """
        Compute guided backprop gradients and returns top layer gradients.

        Parameters:
            input_data : numpy.ndarray or torch.Tensor
                2D for DenseNet, 4D (for 2D images) or 5D (for 3D images) Tensor.
            targets : numpy.ndarray or torch.LongTensor
                1D list of class labels

        Returns:
            gradients : list of numpy.ndarray
                Gradient list of numpy array with same shape as inputs.

        """
        assert isinstance(targets, torch.LongTensor)

        def requires_grad_multidataset(data_list):
            for d in data_list:
                if isinstance(d, list):
                    requires_grad_multidataset(d)
                else:
                    assert isinstance(d, torch.Tensor)
                    d.requires_grad_()

        if isinstance(input_data, list):
            requires_grad_multidataset(input_data)
        else:
            assert isinstance(input_data, torch.Tensor)
            input_data.requires_grad_()

        # Forward pass
        network_output = self.network(input_data)
        # Zero gradients
        self.network.zero_grad()
        # Target for backprop
        one_hot_zeros = torch.zeros(
            network_output.size()[0],
            self.network._num_classes)
        one_hot_output = one_hot_zeros.scatter_(1, targets.unsqueeze(dim=1), 1)
        one_hot_output = one_hot_output.to(self.network.device)
        # Backward pass
        network_output.backward(gradient=one_hot_output)

        def extract_input_gradients_multidataset(input_data):
            for data in input_data:
                if isinstance(data, list):
                    extract_input_gradients_multidataset(data)
                else:
                    self.gradients.append(data.grad.detach().cpu().numpy())

        if isinstance(input_data, list):
            extract_input_gradients_multidataset(input_data)
        else:
            self.gradients.append(input_data.grad.detach().cpu().numpy())

        self._remove_hooks()
        return self.gradients



# coding=utf-8
"""Includes TestMetrics which tests the metrics class of vulcanai."""
import pytest
import numpy as np
import torch
from vulcanai.models.metrics import Metrics
from vulcanai.models.cnn import ConvNet
from torch.utils.data import TensorDataset, DataLoader


# noinspection PyProtectedMember
class TestMetrics:
    """
    Test metric functionality.

    Note: The number of classes NEEDS to stay consistent
    since the metric values were calculated externally.

    """

    @pytest.fixture
    def metrics(self):
        """Create a metric object fixture to test."""
        # TODO: check that this won't mess with replicability
        np.random.seed(1234)
        return Metrics()

    @pytest.fixture
    def cnn_class(self):
        """Create ConvNet with prediction layer."""
        return ConvNet(
            name='Test_ConvNet_class',
            in_dim=(1, 28, 28),
            config={
                'conv_units': [
                    {
                        "in_channels": 10,
                        "out_channels": 16,
                        "kernel_size": (5, 5),
                        "stride": 2
                    },
                    {
                        "in_channels": 16,
                        "out_channels": 1,
                        "kernel_size": (5, 5),
                        "stride": 1,
                        "padding": 2
                    }]
            },
            num_classes=10
        )

    @pytest.fixture
    def cnn_class_binary(self):
        """Create ConvNet with prediction layer."""
        return ConvNet(
            name='Test_ConvNet_class',
            in_dim=(1, 28, 28),
            config={
                'conv_units': [
                    {
                        "in_channels": 10,
                        "out_channels": 16,
                        "kernel_size": (5, 5),
                        "stride": 2
                    },
                    {
                        "in_channels": 16,
                        "out_channels": 1,
                        "kernel_size": (5, 5),
                        "stride": 1,
                        "padding": 2
                    }]
            },
            num_classes=2
        )

    def create_target_predictions(self):
        """Create target and predictions, used in most metric tests."""
        num_items = 300
        test_target = np.random.randint(0, 10, size=num_items)
        test_predictions = np.random.randint(0, 10, size=num_items)

        return test_target, test_predictions

    def test_extract_class_labels(self, metrics):
        """Correctly represents max likelihood class."""
        test_input = np.array([
                [0.2, 0.8],
                [0.7, 0.3],
                [0.25, 0.75]
            ])
        output = metrics.transform_outputs(test_input)
        assert np.all(output == np.array([1, 0, 1]))

    def test_cross_validate_outputs(self, metrics, cnn_class):
        """Tests that the cross-validate outputs are in the correct form."""
        num_items = 300
        test_input = torch.Tensor(np.random.randint(0, 10,
                                                    size=(num_items,
                                                          *cnn_class.in_dim)))
        test_target = torch.LongTensor(np.random.randint(0, 10,
                                                         size=num_items))
        test_dataloader = DataLoader(TensorDataset(test_input, test_target))

        k = 2
        epochs = 2

        averaged_results = metrics.cross_validate(
            cnn_class, test_dataloader, k, epochs,
            average_results=True)
        all_results = metrics.cross_validate(
            cnn_class, test_dataloader, k, epochs,
            average_results=False)

        for k in averaged_results:
            assert isinstance(averaged_results[k], float)

        for k in all_results:
            assert isinstance(all_results[k], list)

    def test_get_score(self, metrics):
        """Test that get score returns correct values, with complex params."""
        test_target, test_predictions = self.create_target_predictions()

        metrics_to_test = ["sensitivity", "specificity", "accuracy"]

        res_dict = metrics.get_score(test_target, test_predictions,
                                     metrics=metrics_to_test, average="macro",
                                     class_converted=True)

        target_sensitivity = 0.08617573063137578
        np.testing.assert_almost_equal(res_dict["sensitivity"],
                                       target_sensitivity)

        target_specificity = 0.89853925
        np.testing.assert_almost_equal(res_dict["specificity"],
                                       target_specificity)

        target_res = 0.08666666666666667
        np.testing.assert_almost_equal(res_dict["accuracy"],
                                       target_res)

    def test_get_confusion_matrix_values(self, metrics):
        """
        Test get_confusion_matrix_values if it returns the correct values.

        The values being true positives, true negatives, false positives,
        and false negatives (tp, tn, fp, fn, respectively). It will return
        class-specific values.

        """
        test_target, test_predictions = self.create_target_predictions()

        tp, tn, fp, fn = metrics.get_confusion_matrix_values(test_target,
                                                             test_predictions)

        target_tp = [2., 4., 3., 0., 2., 3., 1., 5., 3., 3.]
        target_tn = [249., 231., 247., 238., 247., 244., 248., 246., 248.,
                     228.]
        target_fp = [20., 33., 21., 32., 21., 32., 34., 23., 25., 33.]
        target_fn = [29., 32., 29., 30., 30., 21., 17., 26., 24., 36.]

        np.testing.assert_almost_equal(tp, target_tp)
        np.testing.assert_almost_equal(tn, target_tn)
        np.testing.assert_almost_equal(fp, target_fp)
        np.testing.assert_almost_equal(fn, target_fn)

    def test_check_average_parameter(self, metrics):
        """Test the test average-ing parameter returns error as expected."""
        test_target, test_predictions = self.create_target_predictions()

        metrics._check_average_parameter(test_target, test_predictions,
                                         average="macro")
        # try again with binary data
        num_items = 300
        test_target = np.random.randint(0, 2, size=num_items)
        test_predictions = np.random.randint(0, 2, size=num_items)
        with pytest.raises(ValueError):
            metrics._check_average_parameter(test_target, test_predictions,
                                             average="macro")

    def test_get_sensitivity(self, metrics):
        """Test sensitivity metric calculation."""
        test_target, test_predictions = self.create_target_predictions()

        res = metrics.get_sensitivity(test_target, test_predictions,
                                      average="macro")
        target_res = 0.08617573063137578
        np.testing.assert_almost_equal(res, target_res)

        res = metrics.get_sensitivity(test_target, test_predictions)
        target_res = [0.06451613, 0.11111111, 0.09375, 0., 0.0625,
                      0.125, 0.05555556, 0.16129032, 0.11111111, 0.07692308]
        np.testing.assert_almost_equal(res, target_res)

    def test_get_specificity(self, metrics):
        """Test that specificity returns values as expected."""
        test_target, test_predictions = self.create_target_predictions()

        res = metrics.get_specificity(test_target, test_predictions,
                                      average="macro")
        target_res = 0.89853925
        np.testing.assert_almost_equal(res, target_res)

        res = metrics.get_specificity(test_target, test_predictions)
        target_res = [0.92565054, 0.875, 0.92164177, 0.88148147, 0.92164177,
                      0.884058, 0.8794326, 0.91449815, 0.9084249, 0.87356323]
        np.testing.assert_almost_equal(res, target_res)

    def test_get_dice(self, metrics):
        """Test that dic returns values as expected."""
        test_target, test_predictions = self.create_target_predictions()

        res = metrics.get_dice(test_target, test_predictions,
                               average="macro")
        target_res = 0.086294405
        np.testing.assert_almost_equal(res, target_res)

        res = metrics.get_dice(test_target, test_predictions)
        target_res = [0.0754717, 0.10958904, 0.10714286, 0., 0.07272727,
                      0.10169491, 0.03773585, 0.16949153, 0.10909091, 0.08]
        np.testing.assert_almost_equal(res, target_res)

    def test_get_ppv(self, metrics):
        """Test that ppv returns values as expected."""
        test_target, test_predictions = self.create_target_predictions()

        res = metrics.get_ppv(test_target, test_predictions,
                              average="macro")
        target_res = 0.08943070540896628
        np.testing.assert_almost_equal(res, target_res)

        res = metrics.get_ppv(test_target, test_predictions)
        target_res = [0.09090909, 0.10810811, 0.125, 0., 0.08695652,
                      0.08571429, 0.02857143, 0.17857143, 0.10714286,
                      0.08333333]
        np.testing.assert_almost_equal(res, target_res)

    def test_get_npv(self, metrics):
        """Test that npv returns values as expected."""
        test_target, test_predictions = self.create_target_predictions()

        res = metrics.get_npv(test_target, test_predictions,
                              average="macro")
        target_res = 0.8985111
        np.testing.assert_almost_equal(res, target_res)

        res = metrics.get_npv(test_target, test_predictions)
        target_res = [0.89568347, 0.878327, 0.89492756, 0.8880597, 0.89169675,
                      0.92075473, 0.9358491, 0.9044118, 0.9117647, 0.8636364]
        np.testing.assert_almost_equal(res, target_res)

    def test_get_accuracy(self, metrics):
        """Test that accuracy returns values as expected."""
        test_target, test_predictions = self.create_target_predictions()

        res = metrics.get_accuracy(test_target, test_predictions)
        target_res = 0.08666666666666667
        np.testing.assert_almost_equal(res, target_res)

    def test_get_f1(self, metrics):
        """Test that f1 returns values as expected."""
        test_target, test_predictions = self.create_target_predictions()

        res = metrics.get_f1(test_target, test_predictions,
                             average="macro")
        target_res = 0.08629440679047068
        np.testing.assert_almost_equal(res, target_res)

        res = metrics.get_f1(test_target, test_predictions)
        target_res = [0.0754717, 0.10958904, 0.10714286, 0., 0.07272727,
                      0.10169492, 0.03773585, 0.16949153, 0.10909091, 0.08]

        np.testing.assert_almost_equal(res, target_res)

    def test_get_auc(self, metrics):
        """Test that auc returns values as expected."""
        num_items = 300
        test_target = torch.LongTensor(np.random.randint(0, 10,
                                                         size=num_items))

        raw_predictions = np.random.randint(0, 10, size=(300, 10))

        res = metrics.get_auc(test_target, raw_predictions, num_classes=10)

        target_res = [0.40700323779829717, 0.3952020202020202,
                      0.6441231343283582, 0.5640740740740741,
                      0.4857159514925373, 0.5764643719806763,
                      0.4055358550039401, 0.5535435903585562,
                      0.41982091982091985, 0.48482169171824346]

        np.testing.assert_almost_equal(res, target_res)

    def test_run_test(self, metrics, cnn_class):
        """Test that run_test returns values as expected."""
        num_items = 300

        test_input = torch.Tensor(np.random.randint(0, 10,
                                                    size=(num_items,
                                                          *cnn_class.in_dim)))
        test_target = torch.LongTensor(np.random.randint(0, 10,
                                                         size=num_items))
        test_dataloader = DataLoader(TensorDataset(test_input, test_target))

        res_dict = metrics.run_test(cnn_class, test_dataloader)

        required_metrics = ['accuracy', 'macro_sensitivity',
                            'macro_specificity', 'avg_dice', 'macro_ppv',
                            'macro_npv', 'macro_f1', 'macro_auc']

        assert all(k in res_dict for k in required_metrics)

    def test_run_test_binary(self, metrics, cnn_class_binary):
        """Test that run_test returns values as expected."""
        num_items = 300

        test_input = torch.Tensor(np.random.randint(0, 2,
                                                    size=(num_items,
                                                          *cnn_class_binary.in_dim)))
        test_target = torch.LongTensor(np.random.randint(0, 2,
                                                         size=num_items))

        test_dataloader = DataLoader(TensorDataset(test_input, test_target))

        res_dict = metrics.run_test(cnn_class_binary, test_dataloader)

        required_metrics = ['accuracy', 'macro_sensitivity',
                            'macro_specificity', 'avg_dice', 'macro_ppv',
                            'macro_npv', 'macro_f1', 'macro_auc']

        assert all(k in res_dict for k in required_metrics)



# coding=utf-8
"""Defines the DenseNet class."""
import torch
import torch.nn as nn

from .basenetwork import BaseNetwork
from .layers import DenseUnit, FlattenUnit

import logging
from inspect import getfullargspec

from collections import OrderedDict

logger = logging.getLogger(__name__)


class DenseNetConfig:
    """Defines the necessary configuration for a DenseNet."""

    def __init__(self, raw_config):
        """
        Take in user config dict and clean it up.

        Cleaned units is stored in self.units

        Parameters:
            raw_config : dict of dict
                User specified dict

        """
        if 'dense_units' not in raw_config:
            raise KeyError("dense_units must be specified.")

        if not isinstance(raw_config['dense_units'], list):
            raise ValueError("dense_units must be of type list.")

        dense_unit_arg_spec = getfullargspec(DenseUnit)
        dense_unit_arg_spec.args.remove('self')

        # Find the index for where the defaulted values begin
        default_arg_start_index = len(dense_unit_arg_spec.args) - \
            len(dense_unit_arg_spec.defaults)
        default_args = dense_unit_arg_spec.args[default_arg_start_index:]

        # Only look at args that were specified to be overwritten
        override_args = set(default_args).intersection(set(raw_config.keys()))
        # If arguments are specified in lists,
        # check that length corresponds to dense_units
        for arg in override_args:
            if isinstance(raw_config[arg], list):
                if len(raw_config[arg]) != len(raw_config['dense_units']):
                    raise ValueError(
                        "{} list not same length as dense_units.".format(arg))
            else:
                # If a single value is specified, apply to all layers
                raw_config[arg] = [raw_config[arg]] * \
                    len(raw_config['dense_units'])

        # TODO: Think about moving dimension to config file
        _units_per_layer = list([None] + raw_config['dense_units'])
        _unit_pairs = list(zip(_units_per_layer[:-1], _units_per_layer[1:]))

        self.units = []
        for i, (in_feature, out_feature) in enumerate(_unit_pairs):
            temp_unit = {
                'in_features': in_feature,
                'out_features': out_feature
            }
            for arg in override_args:
                temp_unit[arg] = raw_config[arg][i]
            self.units.append(temp_unit)


class DenseNet(BaseNetwork):
    """
    Subclass of BaseNetwork defining a DenseNet.

    Parameters:
        name : str
            The name of the network. Used when saving the file.
        config : dict
            The configuration of the network module, as a dict.
        in_dim : tuple
            The input dimensions of the network. Not required to specify when the
            network has input_networks.
        save_path : str
            The name of the file to which you would like to save this network.
        input_networks : list of BaseNetwork
            A network object provided as input.
        num_classes : int or None
            The number of classes to predict.
        activation : torch.nn.Module
            The desired activation function for use in the network.
        pred_activation : torch.nn.Module
            The desired activation function for use in the prediction layer.
        optim_spec : dict
            A dictionary of parameters for the desired optimizer.
        lr_scheduler : torch.optim.lr_scheduler
            A callable torch.optim.lr_scheduler
        early_stopping : str or None
            So far just 'best_validation_error' is implemented.
        early_stopping_patience: integer
            Number of validation iterations of decreasing loss
            (note -not necessarily every epoch!
            before early stopping is applied.
        early_stopping_metric: string
            Either "loss" or "accuracy" are implemented.
        criter_spec : dict
            criterion specification with name and all its parameters.

    Returns:
        network : DenseNet
            A network of type BaseNetwork.

    """

    def __init__(self, name, config, in_dim=None, save_path=None,
                 input_networks=None, num_classes=None,
                 activation=nn.ReLU(), pred_activation=None,
                 optim_spec={'name': 'Adam', 'lr': 0.001},
                 lr_scheduler=None, early_stopping=None,
                 early_stopping_patience=None,
                 early_stopping_metric="accuracy",
                 criter_spec=nn.CrossEntropyLoss(),
                 device="cuda:0"):
        """Define the DenseNet object."""
        super(DenseNet, self).__init__(
            name, DenseNetConfig(config), in_dim, save_path, input_networks,
            num_classes, activation, pred_activation, optim_spec,
            lr_scheduler, early_stopping, early_stopping_patience,
            early_stopping_metric,
            criter_spec, device)

    def _create_network(self, **kwargs):
        """
        Build the layers of the network into a nn.Sequential object.

        Parameters:
            dense_hid_layers : DenseNetConfig.units (list of dict)
                The hidden layers specification
            activation : torch.nn.Module
                the non-linear activation to apply to each layer

        Returns:
            output : torch.nn.Sequential
                the dense network as a nn.Sequential object

        """
        dense_hid_layers = self._config.units

        if self.input_networks:
            self.in_dim = self._get_in_dim()

        # Build network
        # Specify incoming feature size for the first dense hidden layer
        dense_hid_layers[0]['in_features'] = self.in_dim[0]
        dense_layers = OrderedDict()
        for idx, dense_layer_config in enumerate(dense_hid_layers):
            dense_layer_config['activation'] = kwargs['activation']
            layer_name = 'dense_{}'.format(idx)
            dense_layers[layer_name] = DenseUnit(**dense_layer_config)
        self.network = nn.Sequential(dense_layers)

        if self._num_classes:
            self.network.add_module(
                'classify', DenseUnit(
                    in_features=self._get_out_dim()[0],
                    out_features=self._num_classes,
                    activation=kwargs['pred_activation']))

    def _merge_input_network_outputs(self, tensors):
        output_tensors = [FlattenUnit()(t) for t in tensors]
        return torch.cat(output_tensors, dim=1)

    def __str__(self):
        if self.optim is not None:
            return super(DenseNet, self).__str__() + '\noptim: {}'.format(self.optim)
        else:
            return super(DenseNet, self).__str__()



# -*- coding: utf-8 -*-
"""
This file contains utility methods that many be useful to several dataset
classes.
check_split_ration, stratify, rationed_split, randomshuffler
were all copy-pasted from torchtext because torchtext is not yet packaged
for anaconda and is therefore not yet a reasonable dependency.
See https://github.com/pytorch/text/blob/master/torchtext/data/dataset.py
"""
import pandas as pd
import logging
import copy
import numpy as np

logger = logging.getLogger(__name__)


# TODO: implement
def clean_dataframe(df):
    """
    Goes through and ensures that all nonsensical values are encoded as NaNs
    :param df:
    :return:
    """

    return df


def check_split_ratio(split_ratio):
    """
    Check that the split ratio argument is not malformed

    Parameters:

        split_ratio: desired split ratio, either a list of length 2 or 3
            depending if the validation set is desired.
    
    Returns:
        split ratio as tuple

    """
    valid_ratio = 0.
    if isinstance(split_ratio, float):
        # Only the train set relative ratio is provided
        # Assert in bounds, validation size is zero
        assert 0. < split_ratio < 1., (
            "Split ratio {} not between 0 and 1".format(split_ratio))

        test_ratio = 1. - split_ratio
        return split_ratio, test_ratio, valid_ratio
    elif isinstance(split_ratio, list):
        # A list of relative ratios is provided
        length = len(split_ratio)
        assert length == 2 or length == 3, (
            "Length of split ratio list should be 2 or 3, got {}".format(
                split_ratio))

        # Normalize if necessary
        ratio_sum = sum(split_ratio)
        if not ratio_sum == 1.:
            split_ratio = [float(ratio) / ratio_sum for ratio in split_ratio]

        if length == 2:
            return tuple(split_ratio + [valid_ratio])
        return tuple(split_ratio)
    else:
        raise ValueError('Split ratio must be float or a list, got {}'
                         .format(type(split_ratio)))


def rationed_split(df, train_ratio, test_ratio, validation_ratio):
    """
    Function to split a dataset given ratios. Assumes the ratios given
    are valid (checked using check_split_ratio).

    Parameters:
        df: Dataframe
            The dataframe you want to split
        train_ratio: int
            proportion of the dataset that will go to the train split.
            between 0 and 1
        test_ratio: int
            proportion of the dataset that will go to the test split.
            between 0 and 1
        validation_ratio: int
            proportion of the dataset that will go to the val split.
            between 0 and 1

    Returns:
        indices: tuple of list of indices.
    """
    N = len(df.index)
    perm = np.random.permutation(df.index)
    train_len = int(round(train_ratio * N))

    # Due to possible rounding problems
    if not validation_ratio:
        test_len = N - train_len
    else:
        test_len = int(round(test_ratio * N))

    indices = (perm[:train_len],  # Train
               perm[train_len:train_len + test_len],  # Test
               perm[train_len + test_len:])  # Validation

    return indices


def stitch_datasets(df_main=None, merge_on_columns=None,
                    index_list=None, **dataset_dict):
    """
    Function to produce a single dataset from multiple.

    Parameters:
        df_dict : dictionary of dataframes to concatenated
            dictionary {key = df name: value = dataframe} of dataframes to
            stitch together.
        merge_on_columns : list of strings
            key(s) that specifies which columns to use to uniquely stitch
            dataset (default None)
        index_list: list of strings
            columns to establish as index for final stitched dataset
            (default None)
        dataset_dict : keyword parameter, value is dataframe
            pandas dataframe assigned to keyword argument that produces a
            dictionary variable.

    Returns:
        merged_df : dataframe
            concatenated dataframe

    :Example:

    >>> df_test_one: pd.DataFrame({'A': ['A0', 'A1', 'A2', 'A3'],
                                   'B': ['B0', 'B1', 'B2', 'B3'],
                                   'C': ['C0', 'C1', 'C2', 'C3'],
                                   'D': ['D0', 'D1', 'D2', 'D3']},
                                   index=[0, 1, 2, 3]),

     >>> df_test_two: pd.DataFrame({'A': ['A4', 'A5', 'A6', 'A7'],
                                    'B': ['B4', 'B5', 'B6', 'B7'],
                                    'C': ['C4', 'C5', 'C6', 'C7'],
                                    'D': ['D4', 'D5', 'D6', 'D7']},
                                    index=[4, 5, 6, 7])}
    >>> df_stitched = stitch_datasets(merge_on_columns=['A'], index_list=['A'], df1=df_test_one, df2=df_test_two)

    """
    # Get name of a dataframe to extract
    if df_main is not None:
        merged_df = copy.deepcopy(df_main)
    else:
        first_column = list(dataset_dict)[0]
        merged_df = dataset_dict.pop(first_column)
        merged_df = merged_df.apply(pd.to_numeric, errors='ignore')

    for key in list(dataset_dict):
        logger.info('Combining: {}'.format(key))
        df_two = dataset_dict.pop(key)
        merged_df = pd.concat([merged_df, df_two], sort=False)

    if merge_on_columns is not None:
        # Group by keys, forward fill and backward fill missing data
        # then remove duplicate keys
        merged_df = merged_df.apply(pd.to_numeric, errors='ignore')
        df_groupOn = merged_df.reset_index(drop=True).\
            groupby(merge_on_columns).apply(lambda x: x.bfill().ffill())
        logger.info("\tDropping duplicates")

        #Drop rows where there are duplicates for the merged_on_columns.
        # We first need to dropna based on merged since drop_duplicates
        # ignores null/na values.
        df_groupOn = df_groupOn.dropna(subset=merge_on_columns, how='all')
        df_groupOn = df_groupOn.drop_duplicates(subset=merge_on_columns,
                                                keep='first', inplace=False)
        merged_df = df_groupOn

    if index_list is not None:
        merged_df = merged_df.set_index(index_list, inplace=False)

    logger.info("\nMerge Total columns = {totalCols}, rows = {totalRows} "
        .format(
        totalCols=len(list(merged_df)),
        totalRows=len(merged_df)))
    return merged_df



"""The script to test all layers and SELU activation properties hold."""
import pytest
import numpy as np
import math
import copy
from functools import reduce
import torch
import torch.nn as nn
from torch.nn import modules
from vulcanai.models.layers import BaseUnit, ConvUnit, DenseUnit
from vulcanai.models.utils import selu_weight_init_, selu_bias_init_
from vulcanai.models.dnn import DenseNet
from vulcanai.models.cnn import ConvNet
from torch.utils.data import TensorDataset, DataLoader

torch.manual_seed(1234)

class TestBaseUnit:
    """To test BaseUnit layer."""

    @pytest.fixture
    def baseunit(self):
        """Make Base unit fixture."""
        return BaseUnit()

    def test_init(self, baseunit):
        """Initialization Test of a BaseUnit object."""
        assert isinstance(baseunit, nn.Sequential)
        assert hasattr(baseunit, 'weight_init')
        assert hasattr(baseunit, 'bias_init')
        assert hasattr(baseunit, 'norm')
        assert hasattr(baseunit, 'dropout')
        assert hasattr(baseunit, 'in_shape')
        assert hasattr(baseunit, 'out_shape')
        assert hasattr(baseunit, '_kernel')


class TestDenseUnit:
    """To test DenseUnit layer."""

    @pytest.fixture
    def denseunit(self):
        """Dense unit fixture."""
        return DenseUnit(
            in_features=10,
            out_features=10,
            weight_init=nn.init.xavier_uniform_,
            bias_init=nn.init.zeros_,
            norm='batch',
            activation=nn.ReLU(),
            dropout=0.5
        )

    def test_init(self, denseunit):
        """Initialization Test of a DenseUnit object."""
        assert hasattr(denseunit, 'in_features')
        assert hasattr(denseunit, 'out_features')
        for unit in denseunit.named_children():
            assert any(unit[0] == i for i in ['_kernel', '_norm',
                       '_activation', '_dropout'])
            assert isinstance(unit[1], (nn.Linear,
                              nn.modules.batchnorm._BatchNorm,
                              nn.ReLU, modules.dropout._DropoutNd))
            assert callable(unit[1])

    @pytest.fixture
    def test_denseunit_parameters(self):
        """Create a dictionary with incorrect DenseUnit parameters."""
        return dict(
            in_channels=10,
            out_features=10,
            weight_init=nn.init.xavier_uniform_,
            bias_init=nn.init.zeros_,
            norm='batch',
            activation=nn.ReLU(),
            dropout=0.5
        )

    def test_create_denseunit(self, test_denseunit_parameters):
        """Check if passing wrong parameters raises TypeError."""
        with pytest.raises(TypeError) as e:
            DenseUnit(**test_denseunit_parameters)
        assert 'in_channels' in str(e.value)

    def test_forward(self, denseunit):
        """Confirm size is expected after forward."""
        test_input = torch.rand(size=[10, denseunit.in_features])
        output = denseunit.forward(test_input)
        assert output.shape == torch.rand(size=[10, denseunit.out_features]).shape


class TestConvUnit:
    """To test ConvUnit layers."""

    @pytest.fixture
    def convunit(self):
        """Create ConvUnit fixture."""
        return ConvUnit(
            conv_dim=2,
            in_channels=10,
            out_channels=10,
            kernel_size=(5, 5),
            weight_init=nn.init.xavier_uniform_,
            bias_init=nn.init.zeros_,
            norm='batch',
            activation=nn.ReLU(),
            dropout=0.5
        )

    @pytest.fixture
    def test_convunit_parameters(self):
        """Create dictionary with incorrect ConvUnit parameters."""
        return dict(
            conv_dim=2,
            in_channels=10,
            out_features=10,
            kernel_size=(5, 5),
            weight_init=nn.init.xavier_uniform_,
            bias_init=nn.init.zeros_,
            norm='batch',
            activation=nn.ReLU(),
            dropout=0.5
        )

    def test_create_convunit(self, test_convunit_parameters):
        """Check if passing wrong parameters raises TypeError."""
        with pytest.raises(TypeError) as e:
            ConvUnit(**test_convunit_parameters)
        assert 'out_features' in str(e.value)

    def test_init(self, convunit):
        """Initialization Test of a ConvUnit object."""
        assert hasattr(convunit, 'in_channels')
        assert hasattr(convunit, 'out_channels')
        assert hasattr(convunit, 'kernel_size')
        for unit in convunit.named_children():
            assert any(unit[0] == i for i in
                       ['_kernel', '_norm', '_activation', '_dropout'])
            assert isinstance(unit[1], (modules.conv._ConvNd,
                              nn.modules.batchnorm._BatchNorm,
                              nn.ReLU, modules.dropout._DropoutNd))
            assert callable(unit[1])

    def test_forward(self, convunit):
        """Confirm size is expected after forward."""
        test_input = torch.rand(size=[10, convunit.in_channels, 28, 28])
        output = convunit.forward(test_input)
        # No padding with 2 5x5 kernels leads from 28x28 -> 24x24
        assert output.shape == \
               torch.rand(size=[10, convunit.out_channels, 24, 24]).shape


class TestSeluInit:
    """To test selu initialized layer properties."""

    @pytest.fixture
    def dense_unit(self):
        """Create a dense unit fixture."""
        return DenseUnit(
            in_features=10,
            out_features=10
        )

    @pytest.fixture
    def conv_unit(self):
        """Create a conv unit fixture."""
        return ConvUnit(
            conv_dim=2,
            in_channels=10,
            out_channels=10,
            kernel_size=(5, 5)
        )

    def test_dense_selu_weight_change(self, dense_unit):
        """Confirm SELU weight init properties hold for dense net."""
        starting_weight = copy.deepcopy(dense_unit._kernel.weight)
        fan_in = dense_unit.in_features
        std = round(math.sqrt(1. / fan_in), 1)

        dense_unit.weight_init = selu_weight_init_
        dense_unit._init_weights()
        new_weight = dense_unit._kernel.weight
        assert (torch.equal(starting_weight, new_weight) is False)
        assert (math.isclose(new_weight.std().item(), math.sqrt(1. / fan_in), rel_tol=std) is True)
        assert (int(new_weight.mean().item()) == 0.0)

    def test_conv_selu_weight_change(self, conv_unit):
        """Confirm SELU weight init properties hold for conv net."""
        starting_weight = copy.deepcopy(conv_unit._kernel.weight)
        fan_in = conv_unit._kernel.in_channels * \
            reduce(lambda k1, k2: k1 * k2, conv_unit._kernel.kernel_size)
        std = math.sqrt(1. / fan_in)
        conv_unit.weight_init = selu_weight_init_
        conv_unit._init_weights()
        new_weight = conv_unit._kernel.weight
        assert (torch.equal(starting_weight, new_weight) is False)
        assert (math.isclose(new_weight.std().item(), math.sqrt(1./fan_in), rel_tol=std) is True)
        assert (int(new_weight.mean().item()) == 0)

    def test_dense_selu_bias_change(self, dense_unit):
        """Confirm SELU bias init properties hold for dense net."""
        starting_bias = copy.deepcopy(dense_unit._kernel.bias)

        dense_unit.bias_init = selu_bias_init_
        dense_unit._init_bias()
        new_bias = dense_unit._kernel.bias
        assert (torch.equal(starting_bias, new_bias) is False)
        assert (round(new_bias.std().item(), 1) == 0.0)
        assert (int(new_bias.mean().item()) == 0)

    def test_conv_selu_bias_change(self, conv_unit):
        """Confirm SELU bias init properties hold for conv net."""
        starting_bias = copy.deepcopy(conv_unit._kernel.bias)

        conv_unit.bias_init = selu_bias_init_
        conv_unit._init_bias()
        new_bias = conv_unit._kernel.bias
        assert (torch.equal(starting_bias, new_bias) is False)
        assert (round(new_bias.std().item(), 1) == 0.0)
        assert (int(new_bias.mean().item()) == 0)


class TestSeluInitTrain:
    """To test selu initialization properties hold during training."""

    @pytest.fixture
    def dnn_class(self):
        """Create DenseNet with no prediction layer."""
        return DenseNet(
            name='Test_DenseNet_class',
            in_dim=(200),
            activation=torch.nn.SELU(),
            num_classes=10,
            config={
                'dense_units': [100],
                'dropout': [0.3],
            },
            optim_spec={'name': 'Adam', 'lr': 0.001}
        )

    @pytest.fixture
    def cnn_class(self):
        """Create ConvNet with prediction layer."""
        return ConvNet(
            name='Test_ConvNet_class',
            in_dim=(1, 28, 28),
            activation=torch.nn.SELU(),
            config={
                'conv_units': [
                    {
                        "in_channels": 1,
                        "out_channels": 16,
                        "kernel_size": (5, 5),
                        "stride": 2
                    },
                    {
                        "in_channels": 16,
                        "out_channels": 1,
                        "kernel_size": (5, 5),
                        "stride": 1,
                        "padding": 2
                    }]
            },
            num_classes=10
        )

    def test_selu_trained_dense(self, dnn_class):
        """Confirm SELU weight and bias properties hold for a dense net."""
        fan_in = dnn_class.in_dim[0]
        std = round(math.sqrt(1. / fan_in), 1)
        test_input = torch.rand(size=[10, *dnn_class.in_dim]).float()
        test_target = torch.LongTensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        test_dataloader = DataLoader(TensorDataset(test_input, test_target))
        dnn_class.fit(test_dataloader, test_dataloader, 5)
        trained = dnn_class.network.dense_0._kernel
        assert (round(trained.weight.std().item(), 1) == std)
        assert (int(trained.weight.mean().item()) == 0.0)
        assert (round(trained.bias.std().item(), 1) == 0.0)
        assert (int(trained.bias.mean().item()) == 0.0)

    def test_selu_trained_conv(self, cnn_class):
        """Confirm SELU weight and bias properties hold for a conv net."""
        fan_in = cnn_class.network[0].in_channels * \
            reduce(lambda k1, k2: k1 * k2, cnn_class.network[0].kernel_size)
        std = round(math.sqrt(1. / fan_in), 1)
        test_input = torch.rand(size=[10, *cnn_class.in_dim]).float()
        test_target = torch.LongTensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        test_dataloader = DataLoader(TensorDataset(test_input, test_target))
        cnn_class.fit(test_dataloader, test_dataloader, 5)
        trained = cnn_class.network.conv_0._kernel
        assert (round(trained.weight.std().item(), 1) == std)
        assert (int(trained.weight.mean().item()) == 0.0)
        assert (round(trained.bias.std().item(), 1) == 0.0)
        assert (int(trained.bias.mean().item()) == 0.0)



# -*- coding: utf-8 -*-
""" Imports dataset classes so they can be used directly

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    fashion
    tabulardataset
    multidataset
"""

from .fashion import FashionData
from .tabulardataset import TabularDataset
from .multidataset import MultiDataset

__all__ = [
    'fashion',
    'tabulardataset',
    'utils',
    'FashionData',
    'TabularDataset',
    'MultiDataset'
]

def start():
    " Start module "
